from collections.abc import Collection, Iterable, Sequence
from typing import TypedDict, Any, ClassVar, Final, TypeVar, overload

from django.core.checks.messages import CheckMessage
from django.core.exceptions import MultipleObjectsReturned as BaseMultipleObjectsReturned
from django.core.exceptions import ObjectDoesNotExist, ValidationError
from django.db.models import BaseConstraint, Field, QuerySet
from django.db.models.manager import Manager
from django.db.models.options import Options
from typing_extensions import Required, Unpack, Self
import datetime
from api.models import APIToken, OutboundWebhook
from core.models import ArchiveResult, Snapshot, SnapshotTag, Tag
from django.contrib.admin.models import LogEntry
from django.contrib.auth.models import Group, Permission, User
from django.contrib.contenttypes.models import ContentType
from django.contrib.sessions.models import Session
from django.db.models.expressions import Combinable
from uuid import UUID

class PermissionInitKwargs(TypedDict, total=False):
    id: int | str | Combinable

    name: str | int | Combinable

    content_type: ContentType | Combinable

    codename: str | int | Combinable

class GroupInitKwargs(TypedDict, total=False):
    id: int | str | Combinable

    name: str | int | Combinable

class UserInitKwargs(TypedDict, total=False):
    id: int | str | Combinable

    password: str | int | Combinable

    last_login: str | datetime.datetime | datetime.date | Combinable | None

    is_superuser: bool | Combinable
    """Designates that this user has all permissions without explicitly assigning them."""

    username: str | int | Combinable
    """Required. 150 characters or fewer. Letters, digits and @/./+/-/_ only."""

    first_name: str | int | Combinable

    last_name: str | int | Combinable

    email: str | int | Combinable

    is_staff: bool | Combinable
    """Designates whether the user can log into this admin site."""

    is_active: bool | Combinable
    """Designates whether this user should be treated as active. Unselect this instead of deleting accounts."""

    date_joined: str | datetime.datetime | datetime.date | Combinable

class ContentTypeInitKwargs(TypedDict, total=False):
    id: int | str | Combinable

    app_label: str | int | Combinable

    model: str | int | Combinable

class SessionInitKwargs(TypedDict, total=False):
    session_key: str | int | Combinable

    session_data: str | Combinable

    expire_date: str | datetime.datetime | datetime.date | Combinable

class LogEntryInitKwargs(TypedDict, total=False):
    id: int | str | Combinable

    action_time: str | datetime.datetime | datetime.date | Combinable

    user: User | Combinable

    content_type: ContentType | Combinable | None

    object_id: str | Combinable | None

    object_repr: str | int | Combinable

    action_flag: float | int | str | Combinable

    change_message: str | Combinable

class TagInitKwargs(TypedDict, total=False):
    created_by: User | Combinable

    created: str | datetime.datetime | datetime.date | Combinable

    modified: str | datetime.datetime | datetime.date | Combinable

    old_id: float | int | str | Combinable

    id: str | UUID

    abid: str | int | Combinable | None
    """ABID-format identifier for this entity (e.g. snp_01BJQMF54D093DXEAWZ6JYRPAQ)"""

    name: str | int | Combinable

    slug: str | int | Combinable

class SnapshotTagInitKwargs(TypedDict, total=False):
    id: int | str | Combinable

    snapshot: Snapshot | Combinable

    tag: Tag | Combinable

class SnapshotInitKwargs(TypedDict, total=False):
    created_by: User | Combinable

    created: str | datetime.datetime | datetime.date | Combinable

    modified: str | datetime.datetime | datetime.date | Combinable

    old_id: str | UUID

    id: str | UUID

    abid: str | int | Combinable | None
    """ABID-format identifier for this entity (e.g. snp_01BJQMF54D093DXEAWZ6JYRPAQ)"""

    url: str | int | Combinable

    timestamp: str | int | Combinable

    title: str | int | Combinable | None

    added: str | datetime.datetime | datetime.date | Combinable

    updated: str | datetime.datetime | datetime.date | Combinable | None

class ArchiveResultInitKwargs(TypedDict, total=False):
    created_by: User | Combinable

    created: str | datetime.datetime | datetime.date | Combinable

    modified: str | datetime.datetime | datetime.date | Combinable

    old_id: float | int | str | Combinable

    id: str | UUID

    abid: str | int | Combinable | None
    """ABID-format identifier for this entity (e.g. snp_01BJQMF54D093DXEAWZ6JYRPAQ)"""

    snapshot: Snapshot | Combinable

    extractor: str | int | Combinable

    cmd: Any

    pwd: str | int | Combinable

    cmd_version: str | int | Combinable | None

    output: str | int | Combinable

    start_ts: str | datetime.datetime | datetime.date | Combinable

    end_ts: str | datetime.datetime | datetime.date | Combinable

    status: str | int | Combinable

class APITokenInitKwargs(TypedDict, total=False):
    modified: str | datetime.datetime | datetime.date | Combinable

    id: str | UUID

    abid: str | int | Combinable | None
    """ABID-format identifier for this entity (e.g. snp_01BJQMF54D093DXEAWZ6JYRPAQ)"""

    created_by: User | Combinable

    created: str | datetime.datetime | datetime.date | Combinable

    token: str | int | Combinable

    expires: str | datetime.datetime | datetime.date | Combinable | None

class OutboundWebhookInitKwargs(TypedDict, total=False):
    name: str | int | Combinable
    """Give your webhook a descriptive name (e.g. Notify ACME Slack channel of any new ArchiveResults)."""

    signal: str | int | Combinable
    """The type of event the webhook should fire for (e.g. Create, Update, Delete)."""

    ref: str | int | Combinable
    """Dot import notation of the model the webhook should fire for (e.g. core.models.Snapshot or core.models.ArchiveResult)."""

    endpoint: str | int | Combinable
    """External URL to POST the webhook notification to (e.g. https://someapp.example.com/webhook/some-webhook-receiver)."""

    headers: Any
    """Headers to send with the webhook request."""

    auth_token: str | int | Combinable
    """Authentication token to use in an Authorization header."""

    enabled: bool | Combinable
    """Is this webhook enabled?"""

    keep_last_response: bool | Combinable
    """Should the webhook keep a log of the latest response it got?"""

    updated: str | datetime.datetime | datetime.date | Combinable
    """When the webhook was last updated."""

    last_response: str | int | Combinable
    """Latest response to this webhook."""

    last_success: str | datetime.datetime | datetime.date | Combinable | None
    """When the webhook last succeeded."""

    last_failure: str | datetime.datetime | datetime.date | Combinable | None
    """When the webhook last failed."""

    id: str | UUID

    abid: str | int | Combinable | None
    """ABID-format identifier for this entity (e.g. snp_01BJQMF54D093DXEAWZ6JYRPAQ)"""

    created_by: User | Combinable

    created: str | datetime.datetime | datetime.date | Combinable

    modified: str | datetime.datetime | datetime.date | Combinable

class PermissionInitKwargs(TypedDict, total=False):
    id: int | str | Combinable

    name: str | int | Combinable

    content_type: ContentType | Combinable

    codename: str | int | Combinable

class GroupInitKwargs(TypedDict, total=False):
    id: int | str | Combinable

    name: str | int | Combinable

class UserInitKwargs(TypedDict, total=False):
    id: int | str | Combinable

    password: str | int | Combinable

    last_login: str | datetime.datetime | datetime.date | Combinable | None

    is_superuser: bool | Combinable
    """Designates that this user has all permissions without explicitly assigning them."""

    username: str | int | Combinable
    """Required. 150 characters or fewer. Letters, digits and @/./+/-/_ only."""

    first_name: str | int | Combinable

    last_name: str | int | Combinable

    email: str | int | Combinable

    is_staff: bool | Combinable
    """Designates whether the user can log into this admin site."""

    is_active: bool | Combinable
    """Designates whether this user should be treated as active. Unselect this instead of deleting accounts."""

    date_joined: str | datetime.datetime | datetime.date | Combinable

class ContentTypeInitKwargs(TypedDict, total=False):
    id: int | str | Combinable

    app_label: str | int | Combinable

    model: str | int | Combinable

class SessionInitKwargs(TypedDict, total=False):
    session_key: str | int | Combinable

    session_data: str | Combinable

    expire_date: str | datetime.datetime | datetime.date | Combinable

class LogEntryInitKwargs(TypedDict, total=False):
    id: int | str | Combinable

    action_time: str | datetime.datetime | datetime.date | Combinable

    user: User | Combinable

    content_type: ContentType | Combinable | None

    object_id: str | Combinable | None

    object_repr: str | int | Combinable

    action_flag: float | int | str | Combinable

    change_message: str | Combinable

class TagInitKwargs(TypedDict, total=False):
    created_by: User | Combinable

    created: str | datetime.datetime | datetime.date | Combinable

    modified: str | datetime.datetime | datetime.date | Combinable

    old_id: float | int | str | Combinable

    id: str | UUID

    abid: str | int | Combinable | None
    """ABID-format identifier for this entity (e.g. snp_01BJQMF54D093DXEAWZ6JYRPAQ)"""

    name: str | int | Combinable

    slug: str | int | Combinable

class SnapshotTagInitKwargs(TypedDict, total=False):
    id: int | str | Combinable

    snapshot: Snapshot | Combinable

    tag: Tag | Combinable

class SnapshotInitKwargs(TypedDict, total=False):
    created_by: User | Combinable

    created: str | datetime.datetime | datetime.date | Combinable

    modified: str | datetime.datetime | datetime.date | Combinable

    old_id: str | UUID

    id: str | UUID

    abid: str | int | Combinable | None
    """ABID-format identifier for this entity (e.g. snp_01BJQMF54D093DXEAWZ6JYRPAQ)"""

    url: str | int | Combinable

    timestamp: str | int | Combinable

    title: str | int | Combinable | None

    added: str | datetime.datetime | datetime.date | Combinable

    updated: str | datetime.datetime | datetime.date | Combinable | None

class ArchiveResultInitKwargs(TypedDict, total=False):
    created_by: User | Combinable

    created: str | datetime.datetime | datetime.date | Combinable

    modified: str | datetime.datetime | datetime.date | Combinable

    old_id: float | int | str | Combinable

    id: str | UUID

    abid: str | int | Combinable | None
    """ABID-format identifier for this entity (e.g. snp_01BJQMF54D093DXEAWZ6JYRPAQ)"""

    snapshot: Snapshot | Combinable

    extractor: str | int | Combinable

    cmd: Any

    pwd: str | int | Combinable

    cmd_version: str | int | Combinable | None

    output: str | int | Combinable

    start_ts: str | datetime.datetime | datetime.date | Combinable

    end_ts: str | datetime.datetime | datetime.date | Combinable

    status: str | int | Combinable

class APITokenInitKwargs(TypedDict, total=False):
    modified: str | datetime.datetime | datetime.date | Combinable

    id: str | UUID

    abid: str | int | Combinable | None
    """ABID-format identifier for this entity (e.g. snp_01BJQMF54D093DXEAWZ6JYRPAQ)"""

    created_by: User | Combinable

    created: str | datetime.datetime | datetime.date | Combinable

    token: str | int | Combinable

    expires: str | datetime.datetime | datetime.date | Combinable | None

class OutboundWebhookInitKwargs(TypedDict, total=False):
    name: str | int | Combinable
    """Give your webhook a descriptive name (e.g. Notify ACME Slack channel of any new ArchiveResults)."""

    signal: str | int | Combinable
    """The type of event the webhook should fire for (e.g. Create, Update, Delete)."""

    ref: str | int | Combinable
    """Dot import notation of the model the webhook should fire for (e.g. core.models.Snapshot or core.models.ArchiveResult)."""

    endpoint: str | int | Combinable
    """External URL to POST the webhook notification to (e.g. https://someapp.example.com/webhook/some-webhook-receiver)."""

    headers: Any
    """Headers to send with the webhook request."""

    auth_token: str | int | Combinable
    """Authentication token to use in an Authorization header."""

    enabled: bool | Combinable
    """Is this webhook enabled?"""

    keep_last_response: bool | Combinable
    """Should the webhook keep a log of the latest response it got?"""

    updated: str | datetime.datetime | datetime.date | Combinable
    """When the webhook was last updated."""

    last_response: str | int | Combinable
    """Latest response to this webhook."""

    last_success: str | datetime.datetime | datetime.date | Combinable | None
    """When the webhook last succeeded."""

    last_failure: str | datetime.datetime | datetime.date | Combinable | None
    """When the webhook last failed."""

    id: str | UUID

    abid: str | int | Combinable | None
    """ABID-format identifier for this entity (e.g. snp_01BJQMF54D093DXEAWZ6JYRPAQ)"""

    created_by: User | Combinable

    created: str | datetime.datetime | datetime.date | Combinable

    modified: str | datetime.datetime | datetime.date | Combinable

_Self = TypeVar("_Self", bound=Model)

class ModelStateFieldsCacheDescriptor:
    @overload
    def __get__(self, inst: None, owner: object) -> Self: ...
    @overload
    def __get__(self, inst: object, owner: object) -> dict[Any, Any]: ...

class ModelState:
    db: str | None
    adding: bool
    fields_cache: ModelStateFieldsCacheDescriptor

class ModelBase(type):
    @property
    def _default_manager(cls: type[_Self]) -> Manager[_Self]: ...  # type: ignore[misc]
    @property
    def _base_manager(cls: type[_Self]) -> Manager[_Self]: ...  # type: ignore[misc]

class Model(metaclass=ModelBase):
    # Note: these two metaclass generated attributes don't really exist on the 'Model'
    # class, runtime they are only added on concrete subclasses of 'Model'. The
    # metaclass also sets up correct inheritance from concrete parent models exceptions.
    # Our mypy plugin aligns with this behaviour and will remove the 2 attributes below
    # and re-add them to correct concrete subclasses of 'Model'
    DoesNotExist: Final[type[ObjectDoesNotExist]]
    MultipleObjectsReturned: Final[type[BaseMultipleObjectsReturned]]
    # This 'objects' attribute will be deleted, via the plugin, in favor of managing it
    # to only exist on subclasses it exists on during runtime.
    objects: ClassVar[Manager[Self]]

    _meta: ClassVar[Options[Self]]
    pk: Any
    _state: ModelState
    @overload
    def __init__(self: Permission, **kwargs: Unpack[PermissionInitKwargs]) -> None: ...
    @overload
    def __init__(self: Group, **kwargs: Unpack[GroupInitKwargs]) -> None: ...
    @overload
    def __init__(self: User, **kwargs: Unpack[UserInitKwargs]) -> None: ...
    @overload
    def __init__(self: ContentType, **kwargs: Unpack[ContentTypeInitKwargs]) -> None: ...
    @overload
    def __init__(self: Session, **kwargs: Unpack[SessionInitKwargs]) -> None: ...
    @overload
    def __init__(self: LogEntry, **kwargs: Unpack[LogEntryInitKwargs]) -> None: ...
    @overload
    def __init__(self: Tag, **kwargs: Unpack[TagInitKwargs]) -> None: ...
    @overload
    def __init__(self: SnapshotTag, **kwargs: Unpack[SnapshotTagInitKwargs]) -> None: ...
    @overload
    def __init__(self: Snapshot, **kwargs: Unpack[SnapshotInitKwargs]) -> None: ...
    @overload
    def __init__(self: ArchiveResult, **kwargs: Unpack[ArchiveResultInitKwargs]) -> None: ...
    @overload
    def __init__(self: APIToken, **kwargs: Unpack[APITokenInitKwargs]) -> None: ...
    @overload
    def __init__(self: OutboundWebhook, **kwargs: Unpack[OutboundWebhookInitKwargs]) -> None: ...
    @classmethod
    def add_to_class(cls, name: str, value: Any) -> Any: ...
    @classmethod
    def from_db(cls, db: str | None, field_names: Collection[str], values: Collection[Any]) -> Self: ...
    def _do_update(
        self,
        base_qs: QuerySet[Self],
        using: str | None,
        pk_val: Any,
        values: Collection[tuple[Field, type[Model] | None, Any]],
        update_fields: Iterable[str] | None,
        forced_update: bool,
    ) -> bool: ...
    def delete(self, using: Any = ..., keep_parents: bool = ...) -> tuple[int, dict[str, int]]: ...
    async def adelete(self, using: Any = ..., keep_parents: bool = ...) -> tuple[int, dict[str, int]]: ...
    def full_clean(
        self, exclude: Iterable[str] | None = ..., validate_unique: bool = ..., validate_constraints: bool = ...
    ) -> None: ...
    def clean(self) -> None: ...
    def clean_fields(self, exclude: Collection[str] | None = ...) -> None: ...
    def validate_unique(self, exclude: Collection[str] | None = ...) -> None: ...
    def date_error_message(self, lookup_type: str, field_name: str, unique_for: str) -> ValidationError: ...
    def unique_error_message(self, model_class: type[Self], unique_check: Sequence[str]) -> ValidationError: ...
    def validate_constraints(self, exclude: Collection[str] | None = ...) -> None: ...
    def get_constraints(self) -> list[tuple[type[Model], Sequence[BaseConstraint]]]: ...
    def save(
        self,
        force_insert: bool | tuple[ModelBase, ...] = ...,
        force_update: bool = ...,
        using: str | None = ...,
        update_fields: Iterable[str] | None = ...,
    ) -> None: ...
    async def asave(
        self,
        force_insert: bool | tuple[ModelBase, ...] = ...,
        force_update: bool = ...,
        using: str | None = ...,
        update_fields: Iterable[str] | None = ...,
    ) -> None: ...
    def save_base(
        self,
        raw: bool = ...,
        force_insert: bool | tuple[ModelBase, ...] = ...,
        force_update: bool = ...,
        using: str | None = ...,
        update_fields: Iterable[str] | None = ...,
    ) -> None: ...
    def refresh_from_db(self, using: str | None = ..., fields: Iterable[str] | None = ...) -> None: ...
    async def arefresh_from_db(self, using: str | None = ..., fields: Iterable[str] | None = ...) -> None: ...
    def serializable_value(self, field_name: str) -> Any: ...
    def prepare_database_save(self, field: Field) -> Any: ...
    def get_deferred_fields(self) -> set[str]: ...
    @classmethod
    def check(cls, **kwargs: Any) -> list[CheckMessage]: ...
    def __getstate__(self) -> dict: ...
    def _get_pk_val(self, meta: Options[Self] | None = None) -> str: ...

def model_unpickle(model_id: tuple[str, str] | type[Model]) -> Model: ...



from typing import Any

from django.db.models.expressions import Combinable, Func
from django.db.models.functions.mixins import FixDecimalInputMixin, NumericOutputFieldMixin
from django.db.models.lookups import Transform

class Abs(Transform): ...
class ACos(NumericOutputFieldMixin, Transform): ...
class ASin(NumericOutputFieldMixin, Transform): ...
class ATan(NumericOutputFieldMixin, Transform): ...
class ATan2(NumericOutputFieldMixin, Func): ...
class Ceil(Transform): ...
class Cos(NumericOutputFieldMixin, Transform): ...
class Cot(NumericOutputFieldMixin, Transform): ...
class Degrees(NumericOutputFieldMixin, Transform): ...
class Exp(NumericOutputFieldMixin, Transform): ...
class Floor(Transform): ...
class Ln(NumericOutputFieldMixin, Transform): ...
class Log(FixDecimalInputMixin, NumericOutputFieldMixin, Func): ...
class Mod(FixDecimalInputMixin, NumericOutputFieldMixin, Func): ...
class Pi(NumericOutputFieldMixin, Func): ...
class Power(NumericOutputFieldMixin, Func): ...
class Radians(NumericOutputFieldMixin, Transform): ...
class Random(NumericOutputFieldMixin, Func): ...

class Round(Transform):
    def __init__(self, expression: Combinable | str, precision: int = ..., **extra: Any) -> None: ...

class Sin(NumericOutputFieldMixin, Transform): ...
class Sqrt(NumericOutputFieldMixin, Transform): ...
class Tan(NumericOutputFieldMixin, Transform): ...
class Sign(Transform): ...



import uuid
from functools import wraps
from django.db import connection, transaction
from django.utils import timezone
from huey.exceptions import TaskLockedException

from archivebox.config import CONSTANTS

class SqliteSemaphore:
    def __init__(self, db_path, table_name, name, value=1, timeout=None):
        self.db_path = db_path
        self.table_name = table_name
        self.name = name
        self.value = value
        self.timeout = timeout or 86400  # Set a max age for lock holders

        # Ensure the table exists
        with connection.cursor() as cursor:
            cursor.execute(f"""
                CREATE TABLE IF NOT EXISTS {self.table_name} (
                    id TEXT PRIMARY KEY,
                    name TEXT,
                    timestamp DATETIME
                )
            """)

    def acquire(self, name=None):
        name = name or str(uuid.uuid4())
        now = timezone.now()
        expiration = now - timezone.timedelta(seconds=self.timeout)

        with transaction.atomic():
            # Remove expired locks
            with connection.cursor() as cursor:
                cursor.execute(f"""
                    DELETE FROM {self.table_name}
                    WHERE name = %s AND timestamp < %s
                """, [self.name, expiration])

            # Try to acquire the lock
            with connection.cursor() as cursor:
                cursor.execute(f"""
                    INSERT INTO {self.table_name} (id, name, timestamp)
                    SELECT %s, %s, %s
                    WHERE (
                        SELECT COUNT(*) FROM {self.table_name}
                        WHERE name = %s
                    ) < %s
                """, [name, self.name, now, self.name, self.value])

                if cursor.rowcount > 0:
                    return name

        # If we couldn't acquire the lock, remove our attempted entry
        with connection.cursor() as cursor:
            cursor.execute(f"""
                DELETE FROM {self.table_name}
                WHERE id = %s AND name = %s
            """, [name, self.name])

        return None

    def release(self, name):
        with connection.cursor() as cursor:
            cursor.execute(f"""
                DELETE FROM {self.table_name}
                WHERE id = %s AND name = %s
            """, [name, self.name])
        return cursor.rowcount > 0


LOCKS_DB_PATH = CONSTANTS.DATABASE_FILE.parent / 'locks.sqlite3'


def lock_task_semaphore(db_path, table_name, lock_name, value=1, timeout=None):
    """
    Lock which can be acquired multiple times (default = 1).

    NOTE: no provisions are made for blocking, waiting, or notifying. This is
    just a lock which can be acquired a configurable number of times.

    Example:

    # Allow up to 3 workers to run this task concurrently. If the task is
    # locked, retry up to 2 times with a delay of 60s.
    @huey.task(retries=2, retry_delay=60)
    @lock_task_semaphore('path/to/db.sqlite3', 'semaphore_locks', 'my-lock', 3)
    def my_task():
        ...
    """
    sem = SqliteSemaphore(db_path, table_name, lock_name, value, timeout)
    def decorator(fn):
        @wraps(fn)
        def inner(*args, **kwargs):
            tid = sem.acquire()
            if tid is None:
                raise TaskLockedException(f'unable to acquire lock {lock_name}')
            try:
                return fn(*args, **kwargs)
            finally:
                sem.release(tid)
        return inner
    return decorator



# Generated by Django 5.0.4 on 2024-04-26 05:28

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('api', '0001_initial'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='apitoken',
            options={'verbose_name': 'API Key', 'verbose_name_plural': 'API Keys'},
        ),
    ]



from collections.abc import Iterable, Sequence
from typing import Any

from django.utils.functional import _StrOrPromise

from .message import DEFAULT_ATTACHMENT_MIME_TYPE as DEFAULT_ATTACHMENT_MIME_TYPE
from .message import BadHeaderError as BadHeaderError
from .message import EmailMessage as EmailMessage
from .message import EmailMultiAlternatives as EmailMultiAlternatives
from .message import SafeMIMEMultipart as SafeMIMEMultipart
from .message import SafeMIMEText as SafeMIMEText
from .message import forbid_multi_line_headers as forbid_multi_line_headers
from .utils import DNS_NAME as DNS_NAME
from .utils import CachedDnsName as CachedDnsName

def get_connection(backend: str | None = ..., fail_silently: bool = ..., **kwds: Any) -> Any: ...
def send_mail(
    subject: _StrOrPromise,
    message: _StrOrPromise,
    from_email: str | None,
    recipient_list: Sequence[str],
    fail_silently: bool = ...,
    auth_user: str | None = ...,
    auth_password: str | None = ...,
    connection: Any | None = ...,
    html_message: str | None = ...,
) -> int: ...
def send_mass_mail(
    datatuple: Iterable[tuple[str, str, str | None, list[str]]],
    fail_silently: bool = ...,
    auth_user: str | None = ...,
    auth_password: str | None = ...,
    connection: Any | None = ...,
) -> int: ...
def mail_admins(
    subject: _StrOrPromise,
    message: _StrOrPromise,
    fail_silently: bool = ...,
    connection: Any | None = ...,
    html_message: str | None = ...,
) -> None: ...
def mail_managers(
    subject: _StrOrPromise,
    message: _StrOrPromise,
    fail_silently: bool = ...,
    connection: Any | None = ...,
    html_message: str | None = ...,
) -> None: ...

outbox: list[EmailMessage]



__package__ = 'abx.django'

from django.apps import AppConfig


class ABXConfig(AppConfig):
    name = 'abx'

    def ready(self):
        import abx
        from django.conf import settings
        
        abx.pm.hook.ready(settings=settings)



__package__ = 'archivebox.api'

import secrets
from datetime import timedelta

from django.conf import settings
from django.db import models
from django.utils import timezone

from signal_webhooks.models import WebhookBase

from django_stubs_ext.db.models import TypedModelMeta

from abid_utils.models import ABIDModel, ABIDField, AutoDateTimeField



def generate_secret_token() -> str:
    # returns cryptographically secure string with len() == 32
    return secrets.token_hex(16)


class APIToken(ABIDModel):
    """
    A secret key generated by a User that's used to authenticate REST API requests to ArchiveBox.
    """
    # ABID: apt_<created_ts>_<token_hash>_<user_id_hash>_<uuid_rand>
    abid_prefix = 'apt_'
    abid_ts_src = 'self.created_at'
    abid_uri_src = 'self.created_by_id'
    abid_subtype_src = '"01"'
    abid_rand_src = 'self.id'
    abid_drift_allowed = True

    id = models.UUIDField(primary_key=True, default=None, null=False, editable=False, unique=True, verbose_name='ID')
    abid = ABIDField(prefix=abid_prefix)

    created_by = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE, default=None, null=False)
    created_at = AutoDateTimeField(default=None, null=False, db_index=True)
    modified_at = models.DateTimeField(auto_now=True)

    token = models.CharField(max_length=32, default=generate_secret_token, unique=True)
    expires = models.DateTimeField(null=True, blank=True)

    class Meta(TypedModelMeta):
        verbose_name = "API Key"
        verbose_name_plural = "API Keys"

    def __str__(self) -> str:
        return self.token

    def __repr__(self) -> str:
        return f'<APIToken user={self.created_by.username} token={self.token_redacted}>'

    def __json__(self) -> dict:
        return {
            "TYPE":             "APIToken",    
            "id":               str(self.pk),
            "abid":             str(self.ABID),
            "created_by_id":    str(self.created_by_id),
            "token":            self.token,
            "created_at":       self.created_at.isoformat(),
            "expires":          self.expires_as_iso8601,
        }

    @property
    def expires_as_iso8601(self):
        """Returns the expiry date of the token in ISO 8601 format or a date 100 years in the future if none."""
        expiry_date = self.expires or (timezone.now() + timedelta(days=365 * 100))

        return expiry_date.isoformat()
    
    @property
    def token_redacted(self):
        return f'************{self.token[-4:]}'

    def is_valid(self, for_date=None):
        for_date = for_date or timezone.now()

        if self.expires and self.expires < for_date:
            return False

        return True






# monkey patch django-signals-webhooks to change how it shows up in Admin UI

class OutboundWebhook(ABIDModel, WebhookBase):
    """
    Model used in place of (extending) signals_webhooks.models.WebhookModel. Swapped using:
        settings.SIGNAL_WEBHOOKS_CUSTOM_MODEL = 'api.models.OutboundWebhook'
    """
    abid_prefix = 'whk_'
    abid_ts_src = 'self.created_at'
    abid_uri_src = 'self.endpoint'
    abid_subtype_src = 'self.ref'
    abid_rand_src = 'self.id'
    abid_drift_allowed = True

    id = models.UUIDField(primary_key=True, default=None, null=False, editable=False, unique=True, verbose_name='ID')
    abid = ABIDField(prefix=abid_prefix)

    created_by = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE, default=None, null=False)
    created_at = AutoDateTimeField(default=None, null=False, db_index=True)
    modified_at = models.DateTimeField(auto_now=True)

    # More fields here: WebhookBase...

    WebhookBase._meta.get_field('name').help_text = (
        'Give your webhook a descriptive name (e.g. Notify ACME Slack channel of any new ArchiveResults).')
    WebhookBase._meta.get_field('signal').help_text = (
        'The type of event the webhook should fire for (e.g. Create, Update, Delete).')
    WebhookBase._meta.get_field('ref').help_text = (
        'Dot import notation of the model the webhook should fire for (e.g. core.models.Snapshot or core.models.ArchiveResult).')
    WebhookBase._meta.get_field('endpoint').help_text = (
        'External URL to POST the webhook notification to (e.g. https://someapp.example.com/webhook/some-webhook-receiver).')

    class Meta(WebhookBase.Meta):
        verbose_name = 'API Outbound Webhook'


    def __str__(self) -> str:
        return f'[{self.abid}] {self.ref} -> {self.endpoint}'



from collections.abc import Mapping, Sequence
from typing import Literal, TypeAlias, overload, Any

from django.http.request import HttpRequest
from django.template.exceptions import TemplateDoesNotExist as TemplateDoesNotExist
from django.utils.safestring import SafeString

from .backends.base import _EngineTemplate
DjangoTemplates: TypeAlias = Literal["public_index.html", "base.html", "add.html", "minimal_index.html", "snapshot.html", "index_row.html", "static_index.html", "snapshot_live.html", "progressbar.html", "navigation.html", "actions_as_select.html", "app_index.html", "private_index.html", "login.html", "snapshots_grid.html", "private_index_grid.html", ".DS_Store", "core/public_index.html", "core/base.html", "core/add.html", "core/minimal_index.html", "core/snapshot.html", "core/index_row.html", "core/static_index.html", "core/snapshot_live.html", "core/progressbar.html", "core/navigation.html", "admin/actions_as_select.html", "admin/base.html", "admin/app_index.html", "admin/private_index.html", "admin/login.html", "admin/snapshots_grid.html", "admin/private_index_grid.html", "static/bootstrap.min.css", "static/external.png", "static/favicon.ico", "static/archive.png", "static/add.css", "static/spinner.gif", "static/jquery.min.js", "static/sort_both.png", "static/jquery.dataTables.min.js", "static/admin.css", "static/sort_desc.png", "static/robots.txt", "static/sort_asc.png", "static/jquery.dataTables.min.css", "registration/password_reset_subject.txt", "auth/widgets/read_only_password_hash.html", "registration/password_reset_done.html", "registration/password_change_form.html", "registration/password_reset_email.html", "registration/password_reset_confirm.html", "registration/password_reset_form.html", "registration/logged_out.html", "registration/password_reset_complete.html", "registration/password_change_done.html", "admin/delete_selected_confirmation.html", "admin/change_list_object_tools.html", "admin/delete_confirmation.html", "admin/actions.html", "admin/color_theme_toggle.html", "admin/change_list_results.html", "admin/index.html", "admin/pagination.html", "admin/date_hierarchy.html", "admin/invalid_setup.html", "admin/500.html", "admin/object_history.html", "admin/404.html", "admin/change_list.html", "admin/popup_response.html", "admin/change_form_object_tools.html", "admin/app_list.html", "admin/submit_line.html", "admin/search_form.html", "admin/prepopulated_fields_js.html", "admin/change_form.html", "admin/nav_sidebar.html", "admin/base_site.html", "admin/filter.html", "admin/auth/user/change_password.html", "admin/auth/user/add_form.html", "admin/includes/object_delete_summary.html", "admin/includes/fieldset.html", "admin/edit_inline/stacked.html", "admin/edit_inline/tabular.html", "admin/widgets/many_to_many_raw_id.html", "admin/widgets/radio.html", "admin/widgets/url.html", "admin/widgets/foreign_key_raw_id.html", "admin/widgets/time.html", "admin/widgets/clearable_file_input.html", "admin/widgets/split_datetime.html", "admin/widgets/date.html", "admin/widgets/related_widget_wrapper.html", "django_jsonform/editor.html", "django_jsonform/attrs.html", "admin_data_views/dict_item.html", "admin_data_views/list_item.html", "admin_data_views/admin_data_item_page.html", "admin_data_views/admin_data_table_page.html", "admin_data_views/item.html", "django_extensions/graph_models/original/label.dot", "django_extensions/graph_models/original/relation.dot", "django_extensions/graph_models/original/digraph.dot", "django_extensions/graph_models/django2018/label.dot", "django_extensions/graph_models/django2018/relation.dot", "django_extensions/graph_models/django2018/digraph.dot", "django_extensions/widgets/foreignkey_searchinput.html"]

def get_template(template_name: DjangoTemplates, using: Literal["django"] | None = ...) -> _EngineTemplate: ...
def select_template(template_name_list: list[DjangoTemplates], using: Literal["django"] | None = ...) -> Any: ...
def render_to_string(
    template_name: DjangoTemplates,
    context: Mapping[str, Any] | None = ...,
    request: HttpRequest | None = ...,
    using: Literal["django"] | None = ...,
) -> SafeString: ...



from typing import Any

from django.core.management.base import BaseCommand

class Command(BaseCommand):
    test_runner: Any
    def run_from_argv(self, argv: Any) -> None: ...
    def add_arguments(self, parser: Any) -> None: ...



import datetime
from collections.abc import Iterable, Iterator, Mapping, Sequence
from typing import Any, Literal, Protocol, type_check_only

from django.core.files.base import File
from django.forms.renderers import BaseRenderer
from django.forms.utils import _DataT, _FilesT
from django.utils.choices import _Choices
from django.utils.datastructures import _ListOrTuple
from django.utils.safestring import SafeString
from typing_extensions import Self, TypeAlias

_OptAttrs: TypeAlias = dict[str, Any]

class MediaOrderConflictWarning(RuntimeWarning): ...

class Media:
    def __init__(
        self,
        media: type | None = ...,
        css: dict[str, Sequence[str]] | None = ...,
        js: Sequence[str] | None = ...,
    ) -> None: ...
    def render(self) -> SafeString: ...
    def render_js(self) -> list[SafeString]: ...
    def render_css(self) -> Iterable[SafeString]: ...
    def absolute_path(self, path: str) -> str: ...
    def __getitem__(self, name: str) -> Media: ...
    @staticmethod
    def merge(*lists: Iterable[Any]) -> list[Any]: ...
    def __add__(self, other: Media) -> Media: ...

class MediaDefiningClass(type): ...

class Widget(metaclass=MediaDefiningClass):
    needs_multipart_form: bool
    is_localized: bool
    is_required: bool
    supports_microseconds: bool
    attrs: _OptAttrs
    template_name: str
    def __init__(self, attrs: _OptAttrs | None = ...) -> None: ...
    def __deepcopy__(self, memo: dict[int, Any]) -> Self: ...
    @property
    def is_hidden(self) -> bool: ...
    @property
    def media(self) -> Media: ...
    def subwidgets(self, name: str, value: Any, attrs: _OptAttrs = ...) -> Iterator[dict[str, Any]]: ...
    def format_value(self, value: Any) -> str | None: ...
    def get_context(self, name: str, value: Any, attrs: _OptAttrs | None) -> dict[str, Any]: ...
    def render(
        self, name: str, value: Any, attrs: _OptAttrs | None = ..., renderer: BaseRenderer | None = ...
    ) -> SafeString: ...
    def build_attrs(self, base_attrs: _OptAttrs, extra_attrs: _OptAttrs | None = ...) -> dict[str, Any]: ...
    def value_from_datadict(self, data: _DataT, files: _FilesT, name: str) -> Any: ...
    def value_omitted_from_data(self, data: _DataT, files: _FilesT, name: str) -> bool: ...
    def id_for_label(self, id_: str) -> str: ...
    def use_required_attribute(self, initial: Any) -> bool: ...

class Input(Widget):
    input_type: str
    template_name: str

class TextInput(Input):
    input_type: str
    template_name: str

class NumberInput(Input):
    input_type: str
    template_name: str

class EmailInput(Input):
    input_type: str
    template_name: str

class URLInput(Input):
    input_type: str
    template_name: str

class PasswordInput(Input):
    render_value: bool
    input_type: str
    template_name: str
    def __init__(self, attrs: _OptAttrs | None = ..., render_value: bool = ...) -> None: ...
    def get_context(self, name: str, value: Any, attrs: _OptAttrs | None) -> dict[str, Any]: ...

class HiddenInput(Input):
    choices: _Choices
    input_type: str
    template_name: str

class MultipleHiddenInput(HiddenInput):
    template_name: str

class FileInput(Input):
    allow_multiple_selected: bool
    input_type: str
    template_name: str
    needs_multipart_form: bool
    def format_value(self, value: Any) -> None: ...
    def value_from_datadict(self, data: _DataT, files: _FilesT, name: str) -> Any: ...
    def value_omitted_from_data(self, data: _DataT, files: _FilesT, name: str) -> bool: ...
    def use_required_attribute(self, initial: Any) -> bool: ...

FILE_INPUT_CONTRADICTION: object

class ClearableFileInput(FileInput):
    clear_checkbox_label: str
    initial_text: str
    input_text: str
    template_name: str
    checked: bool
    def clear_checkbox_name(self, name: str) -> str: ...
    def clear_checkbox_id(self, name: str) -> str: ...
    def is_initial(self, value: File | str | None) -> bool: ...
    def get_context(self, name: str, value: Any, attrs: _OptAttrs | None) -> dict[str, Any]: ...
    def value_from_datadict(self, data: _DataT, files: _FilesT, name: str) -> Any: ...
    def value_omitted_from_data(self, data: _DataT, files: _FilesT, name: str) -> bool: ...

class Textarea(Widget):
    template_name: str
    def __init__(self, attrs: _OptAttrs | None = ...) -> None: ...

class DateTimeBaseInput(TextInput):
    format_key: str
    format: str | None
    supports_microseconds: bool
    def __init__(self, attrs: _OptAttrs | None = ..., format: str | None = ...) -> None: ...

class DateInput(DateTimeBaseInput):
    format_key: str
    template_name: str

class DateTimeInput(DateTimeBaseInput):
    format_key: str
    template_name: str

class TimeInput(DateTimeBaseInput):
    format_key: str
    template_name: str

def boolean_check(v: Any) -> bool: ...
@type_check_only
class _CheckCallable(Protocol):
    def __call__(self, value: Any, /) -> bool: ...

class CheckboxInput(Input):
    check_test: _CheckCallable
    input_type: str
    template_name: str
    def __init__(self, attrs: _OptAttrs | None = ..., check_test: _CheckCallable | None = ...) -> None: ...

class ChoiceWidget(Widget):
    allow_multiple_selected: bool
    input_type: str | None
    template_name: str
    option_template_name: str | None
    add_id_index: bool
    checked_attribute: Any
    option_inherits_attrs: bool
    choices: _Choices
    def __init__(self, attrs: _OptAttrs | None = ..., choices: _Choices = ...) -> None: ...
    def subwidgets(self, name: str, value: Any, attrs: _OptAttrs = ...) -> Iterator[dict[str, Any]]: ...
    def options(self, name: str, value: list[str], attrs: _OptAttrs | None = ...) -> Iterator[dict[str, Any]]: ...
    def optgroups(
        self, name: str, value: list[str], attrs: _OptAttrs | None = ...
    ) -> list[tuple[str | None, list[dict[str, Any]], int | None]]: ...
    def get_context(self, name: str, value: Any, attrs: _OptAttrs | None) -> dict[str, Any]: ...
    def create_option(
        self,
        name: str,
        value: Any,
        label: int | str,
        selected: bool,
        index: int,
        subindex: int | None = ...,
        attrs: _OptAttrs | None = ...,
    ) -> dict[str, Any]: ...
    def id_for_label(self, id_: str, index: str = ...) -> str: ...
    def value_from_datadict(self, data: _DataT, files: _FilesT, name: str) -> Any: ...
    def format_value(self, value: Any) -> list[str]: ...  # type: ignore[override]

class Select(ChoiceWidget):
    input_type: str | None
    template_name: str
    option_template_name: str
    add_id_index: bool
    checked_attribute: Any
    option_inherits_attrs: bool
    def get_context(self, name: str, value: Any, attrs: _OptAttrs | None) -> dict[str, Any]: ...
    def use_required_attribute(self, initial: Any) -> bool: ...

class NullBooleanSelect(Select):
    def __init__(self, attrs: _OptAttrs | None = ...) -> None: ...
    def format_value(self, value: Any) -> str: ...  # type: ignore[override]
    def value_from_datadict(self, data: _DataT, files: _FilesT, name: str) -> bool | None: ...

class SelectMultiple(Select):
    allow_multiple_selected: bool
    def value_from_datadict(self, data: _DataT, files: _FilesT, name: str) -> Any: ...
    def value_omitted_from_data(self, data: _DataT, files: _FilesT, name: str) -> bool: ...

class RadioSelect(ChoiceWidget):
    can_add_related: bool
    input_type: str
    template_name: str
    option_template_name: str

class CheckboxSelectMultiple(ChoiceWidget):
    can_add_related: bool
    input_type: str
    template_name: str
    option_template_name: str
    def use_required_attribute(self, initial: Any) -> bool: ...
    def value_omitted_from_data(self, data: _DataT, files: _FilesT, name: str) -> bool: ...
    def id_for_label(self, id_: str, index: str | None = ...) -> str: ...

class MultiWidget(Widget):
    template_name: str
    widgets: Sequence[Widget]
    def __init__(
        self,
        widgets: dict[str, Widget | type[Widget]] | Sequence[Widget | type[Widget]],
        attrs: _OptAttrs | None = ...,
    ) -> None: ...
    @property
    def is_hidden(self) -> bool: ...
    def get_context(self, name: str, value: Any, attrs: _OptAttrs | None) -> dict[str, Any]: ...
    def id_for_label(self, id_: str) -> str: ...
    def value_from_datadict(self, data: _DataT, files: _FilesT, name: str) -> list[Any]: ...
    def value_omitted_from_data(self, data: _DataT, files: _FilesT, name: str) -> bool: ...
    def decompress(self, value: Any) -> Any | None: ...
    @property
    def needs_multipart_form(self) -> bool: ...  # type: ignore[override]

class SplitDateTimeWidget(MultiWidget):
    supports_microseconds: bool
    template_name: str
    widgets: tuple[DateInput, TimeInput]
    def __init__(
        self,
        attrs: _OptAttrs | None = ...,
        date_format: str | None = ...,
        time_format: str | None = ...,
        date_attrs: dict[str, str] | None = ...,
        time_attrs: dict[str, str] | None = ...,
    ) -> None: ...
    def decompress(self, value: Any) -> tuple[datetime.date | None, datetime.time | None]: ...

class SplitHiddenDateTimeWidget(SplitDateTimeWidget):
    template_name: str
    def __init__(
        self,
        attrs: _OptAttrs | None = ...,
        date_format: str | None = ...,
        time_format: str | None = ...,
        date_attrs: dict[str, str] | None = ...,
        time_attrs: dict[str, str] | None = ...,
    ) -> None: ...

class SelectDateWidget(Widget):
    none_value: tuple[Literal[""], str]
    month_field: str
    day_field: str
    year_field: str
    template_name: str
    input_type: str
    select_widget: type[ChoiceWidget]
    date_re: Any
    years: Iterable[int | str]
    months: Mapping[int, str]
    year_none_value: tuple[Literal[""], str]
    month_none_value: tuple[Literal[""], str]
    day_none_value: tuple[Literal[""], str]
    def __init__(
        self,
        attrs: _OptAttrs | None = ...,
        years: Iterable[int | str] | None = ...,
        months: Mapping[int, str] | None = ...,
        empty_label: str | _ListOrTuple[str] | None = ...,
    ) -> None: ...
    def get_context(self, name: str, value: Any, attrs: _OptAttrs | None) -> dict[str, Any]: ...
    def format_value(self, value: Any) -> dict[str, str | int | None]: ...  # type: ignore[override]
    def id_for_label(self, id_: str) -> str: ...
    def value_from_datadict(self, data: _DataT, files: _FilesT, name: str) -> str | None | Any: ...
    def value_omitted_from_data(self, data: _DataT, files: _FilesT, name: str) -> bool: ...



# Generated by Django 5.1.1 on 2024-10-03 09:50

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("machine", "0003_alter_installedbinary_options_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="installedbinary",
            name="abspath",
            field=models.CharField(blank=True, default=None, max_length=255),
        ),
        migrations.AlterField(
            model_name="installedbinary",
            name="binprovider",
            field=models.CharField(blank=True, default=None, max_length=31),
        ),
        migrations.AlterField(
            model_name="installedbinary",
            name="machine",
            field=models.ForeignKey(
                blank=True,
                default=None,
                on_delete=django.db.models.deletion.CASCADE,
                to="machine.machine",
            ),
        ),
        migrations.AlterField(
            model_name="installedbinary",
            name="name",
            field=models.CharField(blank=True, default=None, max_length=63),
        ),
        migrations.AlterField(
            model_name="installedbinary",
            name="sha256",
            field=models.CharField(blank=True, default=None, max_length=64),
        ),
        migrations.AlterField(
            model_name="installedbinary",
            name="version",
            field=models.CharField(blank=True, default=None, max_length=32),
        ),
    ]



from typing import Any, Literal, overload

from django.contrib.gis.gdal.raster.base import GDALRasterBase

class GDALBand(GDALRasterBase):
    source: Any
    def __init__(self, source: Any, index: Any) -> None: ...
    @property
    def description(self) -> str: ...
    @property
    def width(self) -> int: ...
    @property
    def height(self) -> int: ...
    @property
    def pixel_count(self) -> int: ...
    def statistics(
        self, refresh: bool = ..., approximate: bool = ...
    ) -> tuple[float, float, float, float] | tuple[None, None, None, None]: ...
    @property
    def min(self) -> float | None: ...
    @property
    def max(self) -> float | None: ...
    @property
    def mean(self) -> float | None: ...
    @property
    def std(self) -> float | None: ...
    @property
    def nodata_value(self) -> float | None: ...
    @nodata_value.setter
    def nodata_value(self, value: float | None) -> None: ...
    @overload
    def datatype(self, as_string: Literal[False] = ...) -> int: ...
    @overload
    def datatype(self, as_string: Literal[True]) -> str: ...
    @overload
    def color_interp(self, as_string: Literal[False] = ...) -> int: ...
    @overload
    def color_interp(self, as_string: Literal[True]) -> str: ...
    def data(
        self,
        data: Any | None = ...,
        offset: Any | None = ...,
        size: Any | None = ...,
        shape: Any | None = ...,
        as_memoryview: bool = ...,
    ) -> Any: ...

class BandList(list[GDALBand]):
    source: Any
    def __init__(self, source: Any) -> None: ...



from django.contrib import messages

from archivebox.search import query_search_index

class SearchResultsAdminMixin:
    def get_search_results(self, request, queryset, search_term: str):
        """Enhances the search queryset with results from the search backend"""
        
        qs, use_distinct = super().get_search_results(request, queryset, search_term)

        search_term = search_term.strip()
        if not search_term:
            return qs.distinct(), use_distinct
        try:
            qsearch = query_search_index(search_term)
            qs = qs | qsearch
        except Exception as err:
            print(f'[!] Error while using search backend: {err.__class__.__name__} {err}')
            messages.add_message(request, messages.WARNING, f'Error from the search backend, only showing results from default admin search fields - Error: {err}')
        
        return qs.distinct(), use_distinct



__package__ = 'archivebox.misc'

# TODO: merge/dedupe this file with archivebox/logging_util.py


import sys
from typing import Optional, Union, Tuple, List
from collections import defaultdict
from random import randint

from benedict import benedict
from rich.console import Console
from rich.highlighter import Highlighter

# SETUP RICH CONSOLE / TTY detection / COLOR / PROGRESS BARS
CONSOLE = Console()
STDERR = Console(stderr=True)
IS_TTY = CONSOLE.is_interactive

class RainbowHighlighter(Highlighter):
    def highlight(self, text):
        for index in range(len(text)):
            text.stylize(f"color({randint(90, 98)})", index, index + 1)

rainbow = RainbowHighlighter()


DEFAULT_CLI_COLORS = benedict(
    {
        "reset": "\033[00;00m",
        "lightblue": "\033[01;30m",
        "lightyellow": "\033[01;33m",
        "lightred": "\033[01;35m",
        "red": "\033[01;31m",
        "green": "\033[01;32m",
        "blue": "\033[01;34m",
        "white": "\033[01;37m",
        "black": "\033[01;30m",
    }
)
ANSI = benedict({k: '' for k in DEFAULT_CLI_COLORS.keys()})

COLOR_DICT = defaultdict(lambda: [(0, 0, 0), (0, 0, 0)], {
    '00': [(0, 0, 0), (0, 0, 0)],
    '30': [(0, 0, 0), (0, 0, 0)],
    '31': [(255, 0, 0), (128, 0, 0)],
    '32': [(0, 200, 0), (0, 128, 0)],
    '33': [(255, 255, 0), (128, 128, 0)],
    '34': [(0, 0, 255), (0, 0, 128)],
    '35': [(255, 0, 255), (128, 0, 128)],
    '36': [(0, 255, 255), (0, 128, 128)],
    '37': [(255, 255, 255), (255, 255, 255)],
})

# Logging Helpers (DEPRECATED, use rich.print instead going forward)
def stdout(*args, color: Optional[str]=None, prefix: str='', config: Optional[benedict]=None) -> None:
    ansi = DEFAULT_CLI_COLORS if (config or {}).get('USE_COLOR') else ANSI

    if color:
        strs = [ansi[color], ' '.join(str(a) for a in args), ansi['reset'], '\n']
    else:
        strs = [' '.join(str(a) for a in args), '\n']

    sys.stdout.write(prefix + ''.join(strs))

def stderr(*args, color: Optional[str]=None, prefix: str='', config: Optional[benedict]=None) -> None:
    ansi = DEFAULT_CLI_COLORS if (config or {}).get('USE_COLOR') else ANSI

    if color:
        strs = [ansi[color], ' '.join(str(a) for a in args), ansi['reset'], '\n']
    else:
        strs = [' '.join(str(a) for a in args), '\n']

    sys.stderr.write(prefix + ''.join(strs))

def hint(text: Union[Tuple[str, ...], List[str], str], prefix='    ', config: Optional[benedict]=None) -> None:
    ansi = DEFAULT_CLI_COLORS if (config or {}).get('USE_COLOR') else ANSI

    if isinstance(text, str):
        stderr('{}{lightred}Hint:{reset} {}'.format(prefix, text, **ansi))
    else:
        stderr('{}{lightred}Hint:{reset} {}'.format(prefix, text[0], **ansi))
        for line in text[1:]:
            stderr('{}      {}'.format(prefix, line))



from typing import IO

from django.core.files.base import File
from typing_extensions import Self

class UploadedFile(File):
    content_type: str | None
    charset: str | None
    content_type_extra: dict[str, str] | None
    size: int | None  # type: ignore[assignment]
    name: str | None
    def __init__(
        self,
        file: IO | None = ...,
        name: str | None = ...,
        content_type: str | None = ...,
        size: int | None = ...,
        charset: str | None = ...,
        content_type_extra: dict[str, str] | None = ...,
    ) -> None: ...

class TemporaryUploadedFile(UploadedFile):
    def __init__(
        self,
        name: str,
        content_type: str | None,
        size: int | None,
        charset: str | None,
        content_type_extra: dict[str, str] | None = ...,
    ) -> None: ...
    def temporary_file_path(self) -> str: ...

class InMemoryUploadedFile(UploadedFile):
    field_name: str | None
    def __init__(
        self,
        file: IO,
        field_name: str | None,
        name: str | None,
        content_type: str | None,
        size: int | None,
        charset: str | None,
        content_type_extra: dict[str, str] | None = ...,
    ) -> None: ...
    def open(self, mode: str | None = ...) -> Self: ...  # type: ignore[override]

class SimpleUploadedFile(InMemoryUploadedFile):
    def __init__(self, name: str, content: bytes | None, content_type: str = ...) -> None: ...
    @classmethod
    def from_dict(cls, file_dict: dict[str, str | bytes]) -> Self: ...



from collections.abc import Iterable, Sequence
from typing import Any, TypeVar

from _typeshed import Unused
from django.core.validators import _ValidatorCallable
from django.db.backends.base.base import BaseDatabaseWrapper
from django.db.models import Field
from django.db.models.expressions import Combinable, Expression
from django.db.models.fields import NOT_PROVIDED, _ErrorMessagesDict, _ErrorMessagesMapping
from django.db.models.fields.mixins import CheckFieldDefaultMixin
from django.db.models.lookups import Transform
from django.utils.choices import _Choices
from django.utils.functional import _StrOrPromise

# __set__ value type
_ST = TypeVar("_ST")
# __get__ return type
_GT = TypeVar("_GT")

class ArrayField(CheckFieldDefaultMixin, Field[_ST, _GT]):
    _pyi_private_set_type: Sequence[Any] | Combinable
    _pyi_private_get_type: list[Any]

    empty_strings_allowed: bool
    default_error_messages: _ErrorMessagesDict
    base_field: Field
    size: int | None
    default_validators: Sequence[_ValidatorCallable]
    from_db_value: Any
    def __init__(
        self,
        base_field: Field,
        size: int | None = ...,
        *,
        verbose_name: _StrOrPromise | None = ...,
        name: str | None = ...,
        primary_key: bool = ...,
        max_length: int | None = ...,
        unique: bool = ...,
        blank: bool = ...,
        null: bool = ...,
        db_index: bool = ...,
        default: Any = ...,
        db_default: type[NOT_PROVIDED] | Expression | _ST = ...,
        editable: bool = ...,
        auto_created: bool = ...,
        serialize: bool = ...,
        unique_for_date: str | None = ...,
        unique_for_month: str | None = ...,
        unique_for_year: str | None = ...,
        choices: _Choices | None = ...,
        help_text: _StrOrPromise = ...,
        db_column: str | None = ...,
        db_comment: str | None = ...,
        db_tablespace: str | None = ...,
        validators: Iterable[_ValidatorCallable] = ...,
        error_messages: _ErrorMessagesMapping | None = ...,
    ) -> None: ...
    @property
    def description(self) -> str: ...  # type: ignore[override]
    def cast_db_type(self, connection: BaseDatabaseWrapper) -> str: ...
    def get_placeholder(self, value: Unused, compiler: Unused, connection: BaseDatabaseWrapper) -> str: ...
    def get_transform(self, name: str) -> type[Transform] | None: ...



__package__ = 'archivebox.queues'

import time
import signal
import psutil
import shutil
import subprocess
from pathlib import Path
from rich import print

from typing import Dict, cast

from supervisor.xmlrpc import SupervisorTransport
from xmlrpc.client import ServerProxy

from archivebox.config.permissions import ARCHIVEBOX_USER

from .settings import SUPERVISORD_CONFIG_FILE, DATA_DIR, PID_FILE, get_sock_file, LOG_FILE, WORKERS_DIR, TMP_DIR, LOGS_DIR

from typing import Iterator

def follow(file, sleep_sec=0.1) -> Iterator[str]:
    """ Yield each line from a file as they are written.
    `sleep_sec` is the time to sleep after empty reads. """
    line = ''
    while True:
        tmp = file.readline()
        if tmp is not None and tmp != "":
            line += tmp
            if line.endswith("\n"):
                yield line
                line = ''
        elif sleep_sec:
            time.sleep(sleep_sec)


def create_supervisord_config():
    config_content = f"""
[supervisord]
nodaemon = true
environment = IS_SUPERVISORD_PARENT="true"
pidfile = {TMP_DIR}/{PID_FILE.name}
logfile = {LOGS_DIR}/{LOG_FILE.name}
childlogdir = {LOGS_DIR}
directory = {DATA_DIR}
strip_ansi = true
nocleanup = true
user = {ARCHIVEBOX_USER}

[unix_http_server]
file = {get_sock_file()}
chmod = 0700

[supervisorctl]
serverurl = unix://{get_sock_file()}

[rpcinterface:supervisor]
supervisor.rpcinterface_factory = supervisor.rpcinterface:make_main_rpcinterface

[include]
files = {WORKERS_DIR}/*.conf

"""
    SUPERVISORD_CONFIG_FILE.write_text(config_content)

def create_worker_config(daemon):
    Path.mkdir(WORKERS_DIR, exist_ok=True)
    
    name = daemon['name']
    configfile = WORKERS_DIR / f"{name}.conf"

    config_content = f"[program:{name}]\n"
    for key, value in daemon.items():
        if key == 'name':
            continue
        config_content += f"{key}={value}\n"
    config_content += "\n"

    configfile.write_text(config_content)


def get_existing_supervisord_process():
    try:
        transport = SupervisorTransport(None, None, f"unix://{get_sock_file()}")
        server = ServerProxy("http://localhost", transport=transport)
        current_state = cast(Dict[str, int | str], server.supervisor.getState())
        if current_state["statename"] == "RUNNING":
            pid = server.supervisor.getPID()
            print(f"[🦸‍♂️] Supervisord connected (pid={pid}) via unix://{str(get_sock_file()).replace(str(DATA_DIR), '.')}.")
            return server.supervisor
    except FileNotFoundError:
        return None
    except Exception as e:
        print(f"Error connecting to existing supervisord: {str(e)}")
        return None

def stop_existing_supervisord_process():
    try:
        pid = int(PID_FILE.read_text())
    except FileNotFoundError:
        return
    except ValueError:
        PID_FILE.unlink()
        return

    try:
        print(f"[🦸‍♂️] Stopping supervisord process (pid={pid})...")
        proc = psutil.Process(pid)
        proc.terminate()
        proc.wait()
    except Exception:
        pass
    try:
        PID_FILE.unlink()
    except FileNotFoundError:
        pass

def start_new_supervisord_process(daemonize=False):
    print(f"[🦸‍♂️] Supervisord starting{' in background' if daemonize else ''}...")
    # Create a config file in the current working directory
    
    # clear out existing stale state files
    shutil.rmtree(WORKERS_DIR, ignore_errors=True)
    PID_FILE.unlink(missing_ok=True)
    get_sock_file().unlink(missing_ok=True)
    SUPERVISORD_CONFIG_FILE.unlink(missing_ok=True)
    
    create_supervisord_config()

    # Start supervisord
    subprocess.Popen(
        f"supervisord --configuration={SUPERVISORD_CONFIG_FILE}",
        stdin=None,
        shell=True,
        start_new_session=daemonize,
    )

    def exit_signal_handler(signum, frame):
        if signum != 13:
            print(f"\n[🦸‍♂️] Supervisord got stop signal ({signal.strsignal(signum)}). Terminating child processes...")
        stop_existing_supervisord_process()
        raise SystemExit(0)

    # Monitor for termination signals and cleanup child processes
    if not daemonize:
        signal.signal(signal.SIGINT, exit_signal_handler)
        signal.signal(signal.SIGHUP, exit_signal_handler)
        signal.signal(signal.SIGPIPE, exit_signal_handler)
        signal.signal(signal.SIGTERM, exit_signal_handler)
    # otherwise supervisord will containue in background even if parent proc is ends (aka daemon mode)

    time.sleep(2)

    return get_existing_supervisord_process()

def get_or_create_supervisord_process(daemonize=False):
    supervisor = get_existing_supervisord_process()
    if supervisor is None:
        stop_existing_supervisord_process()
        supervisor = start_new_supervisord_process(daemonize=daemonize)
        time.sleep(0.5)

    assert supervisor, "Failed to start supervisord or connect to it!"
    supervisor.getPID()  # make sure it doesn't throw an exception
    
    return supervisor

def start_worker(supervisor, daemon, lazy=False):
    assert supervisor.getPID()

    print(f"[🦸‍♂️] Supervisord starting new subprocess worker: {daemon['name']}...")
    create_worker_config(daemon)

    result = supervisor.reloadConfig()
    added, changed, removed = result[0]
    # print(f"Added: {added}, Changed: {changed}, Removed: {removed}")
    for removed in removed:
        supervisor.stopProcessGroup(removed)
        supervisor.removeProcessGroup(removed)
    for changed in changed:
        supervisor.stopProcessGroup(changed)
        supervisor.removeProcessGroup(changed)
        supervisor.addProcessGroup(changed)
    for added in added:
        supervisor.addProcessGroup(added)

    time.sleep(1)

    for _ in range(10):
        procs = supervisor.getAllProcessInfo()
        for proc in procs:
            if proc['name'] == daemon["name"]:
                # See process state diagram here: http://supervisord.org/subprocess.html
                if proc['statename'] == 'RUNNING':
                    print(f"     - Worker {daemon['name']}: already {proc['statename']} ({proc['description']})")
                    return proc
                else:
                    if not lazy:
                        supervisor.startProcessGroup(daemon["name"], True)
                    proc = supervisor.getProcessInfo(daemon["name"])
                    print(f"     - Worker {daemon['name']}: started {proc['statename']} ({proc['description']})")
                return proc

        # retry in a second in case it's slow to launch
        time.sleep(0.5)

    raise Exception(f"Failed to start worker {daemon['name']}! Only found: {procs}")


def watch_worker(supervisor, daemon_name, interval=5):
    """loop continuously and monitor worker's health"""
    while True:
        proc = get_worker(supervisor, daemon_name)
        if not proc:
            raise Exception("Worker dissapeared while running! " + daemon_name)

        if proc['statename'] == 'STOPPED':
            return proc

        if proc['statename'] == 'RUNNING':
            time.sleep(1)
            continue

        if proc['statename'] in ('STARTING', 'BACKOFF', 'FATAL', 'EXITED', 'STOPPING'):
            print(f'[🦸‍♂️] WARNING: Worker {daemon_name} {proc["statename"]} {proc["description"]}')
            time.sleep(interval)
            continue

def tail_worker_logs(log_path: str):
    get_or_create_supervisord_process(daemonize=False)
    
    from rich.live import Live
    from rich.table import Table
    
    table = Table()
    table.add_column("TS")
    table.add_column("URL")
    
    try:
        with Live(table, refresh_per_second=1) as live:  # update 4 times a second to feel fluid
            with open(log_path, 'r') as f:
                for line in follow(f):
                    if '://' in line:
                        live.console.print(f"Working on: {line.strip()}")
                    table.add_row("123124234", line.strip())
    except KeyboardInterrupt:
        print("\n[🛑] Got Ctrl+C, stopping gracefully...")
    except SystemExit:
        pass

def get_worker(supervisor, daemon_name):
    try:
        return supervisor.getProcessInfo(daemon_name)
    except Exception:
        pass
    return None

def stop_worker(supervisor, daemon_name):
    proc = get_worker(supervisor, daemon_name)

    for _ in range(10):
        if not proc:
            # worker does not exist (was never running or configured in the first place)
            return True
        
        # See process state diagram here: http://supervisord.org/subprocess.html
        if proc['statename'] == 'STOPPED':
            # worker was configured but has already stopped for some reason
            supervisor.removeProcessGroup(daemon_name)
            return True
        else:
            # worker was configured and is running, stop it now
            supervisor.stopProcessGroup(daemon_name)

        # wait 500ms and then re-check to make sure it's really stopped
        time.sleep(0.5)
        proc = get_worker(supervisor, daemon_name)

    raise Exception(f"Failed to stop worker {daemon_name}!")




def start_server_workers(host='0.0.0.0', port='8000', daemonize=False):
    supervisor = get_or_create_supervisord_process(daemonize=daemonize)
    
    bg_workers = [
        {
            "name": "worker_scheduler",
            "command": "archivebox manage djangohuey --queue system_tasks -w 4 -k thread --disable-health-check --flush-locks",
            "autostart": "true",
            "autorestart": "true",
            "stdout_logfile": "logs/worker_scheduler.log",
            "redirect_stderr": "true",
        },
        {
            "name": "worker_system_tasks",
            "command": "archivebox manage djangohuey --queue system_tasks -w 4 -k thread --no-periodic --disable-health-check",
            "autostart": "true",
            "autorestart": "true",
            "stdout_logfile": "logs/worker_system_tasks.log",
            "redirect_stderr": "true",
        },
    ]
    fg_worker = {
        "name": "worker_daphne",
        "command": f"daphne --bind={host} --port={port} --application-close-timeout=600 archivebox.core.asgi:application",
        "autostart": "false",
        "autorestart": "true",
        "stdout_logfile": "logs/worker_daphne.log",
        "redirect_stderr": "true",
    }

    print()
    start_worker(supervisor, fg_worker)
    print()
    for worker in bg_workers:
        start_worker(supervisor, worker)
    print()

    if not daemonize:
        try:
            watch_worker(supervisor, "worker_daphne")
        except KeyboardInterrupt:
            print("\n[🛑] Got Ctrl+C, stopping gracefully...")
        except SystemExit:
            pass
        except BaseException as e:
            print(f"\n[🛑] Got {e.__class__.__name__} exception, stopping web server gracefully...")
            raise
        finally:
            stop_worker(supervisor, "worker_daphne")
            time.sleep(0.5)


def start_cli_workers(watch=False):
    supervisor = get_or_create_supervisord_process(daemonize=False)
    
    fg_worker = {
        "name": "worker_system_tasks",
        "command": "archivebox manage djangohuey --queue system_tasks",
        "autostart": "true",
        "autorestart": "true",
        "stdout_logfile": "logs/worker_system_tasks.log",
        "redirect_stderr": "true",
    }

    start_worker(supervisor, fg_worker)

    if watch:
        try:
            watch_worker(supervisor, "worker_system_tasks")
        except KeyboardInterrupt:
            print("\n[🛑] Got Ctrl+C, stopping gracefully...")
        except SystemExit:
            pass
        except BaseException as e:
            print(f"\n[🛑] Got {e.__class__.__name__} exception, stopping web server gracefully...")
            raise
        finally:
            stop_worker(supervisor, "worker_system_tasks")
            stop_worker(supervisor, "worker_scheduler")
            time.sleep(0.5)
    return fg_worker


# def main(daemons):
#     supervisor = get_or_create_supervisord_process(daemonize=False)

#     worker = start_worker(supervisor, daemons["webworker"])
#     pprint(worker)

#     print("All processes started in background.")
    
    # Optionally you can block the main thread until an exit signal is received:
    # try:
    #     signal.pause()
    # except KeyboardInterrupt:
    #     pass
    # finally:
    #     stop_existing_supervisord_process()

# if __name__ == "__main__":

#     DAEMONS = {
#         "webworker": {
#             "name": "webworker",
#             "command": "python3 -m http.server 9000",
#             "directory": str(cwd),
#             "autostart": "true",
#             "autorestart": "true",
#             "stdout_logfile": cwd / "webworker.log",
#             "stderr_logfile": cwd / "webworker_error.log",
#         },
#     }
#     main(DAEMONS, cwd)



import datetime
from collections.abc import Iterable, Mapping
from io import BytesIO
from re import Pattern
from typing import Any, Awaitable, BinaryIO, Callable, Literal, NoReturn, TypeVar, overload, type_check_only

from django.contrib.auth.base_user import AbstractBaseUser
from django.contrib.auth.models import AnonymousUser
from django.contrib.sessions.backends.base import SessionBase
from django.contrib.sites.models import Site
from django.core.files import uploadedfile, uploadhandler
from django.urls import ResolverMatch
from django.utils.datastructures import CaseInsensitiveMapping, ImmutableList, MultiValueDict
from django.utils.functional import cached_property
from typing_extensions import Self, TypeAlias

RAISE_ERROR: object
host_validation_re: Pattern[str]

class UnreadablePostError(OSError): ...
class RawPostDataException(Exception): ...

_UploadHandlerList: TypeAlias = list[uploadhandler.FileUploadHandler] | ImmutableList[uploadhandler.FileUploadHandler]

class HttpHeaders(CaseInsensitiveMapping[str]):
    HTTP_PREFIX: str
    UNPREFIXED_HEADERS: set[str]
    def __init__(self, environ: Mapping[str, Any]) -> None: ...
    @classmethod
    def parse_header_name(cls, header: str) -> str | None: ...
    @classmethod
    def to_wsgi_name(cls, header: str) -> str: ...
    @classmethod
    def to_asgi_name(cls, header: str) -> str: ...
    @classmethod
    def to_wsgi_names(cls, headers: Mapping[str, Any]) -> dict[str, Any]: ...
    @classmethod
    def to_asgi_names(cls, headers: Mapping[str, Any]) -> dict[str, Any]: ...

class HttpRequest(BytesIO):
    GET: _ImmutableQueryDict
    POST: _ImmutableQueryDict
    COOKIES: dict[str, str]
    META: dict[str, Any]
    FILES: MultiValueDict[str, uploadedfile.UploadedFile]
    path: str
    path_info: str
    method: str | None
    resolver_match: ResolverMatch | None
    content_type: str | None
    content_params: dict[str, str] | None
    _body: bytes
    _stream: BinaryIO
    # Attributes added by optional parts of Django
    # django.contrib.admin views:
    current_app: str
    # django.contrib.auth.middleware.AuthenticationMiddleware:
    user: AbstractBaseUser | AnonymousUser
    # django.contrib.auth.middleware.AuthenticationMiddleware:
    auser: Callable[[], Awaitable[AbstractBaseUser | AnonymousUser]]
    # django.middleware.locale.LocaleMiddleware:
    LANGUAGE_CODE: str
    # django.contrib.sites.middleware.CurrentSiteMiddleware
    site: Site
    # django.contrib.sessions.middleware.SessionMiddleware
    session: SessionBase
    # The magic. If we instantiate HttpRequest directly somewhere, it has
    # mutable GET and POST. However, both ASGIRequest and WSGIRequest have immutable,
    # so when we use HttpRequest to refer to any of them we want exactly this.
    # Case when some function creates *exactly* HttpRequest (not subclass)
    # remain uncovered, however it's probably the best solution we can afford.
    def __new__(cls) -> _MutableHttpRequest: ...
    # When both __init__ and __new__ are present, mypy will prefer __init__
    # (see comments in mypy.checkmember.type_object_type)
    # def __init__(self) -> None: ...
    def get_host(self) -> str: ...
    def get_port(self) -> str: ...
    def get_full_path(self, force_append_slash: bool = ...) -> str: ...
    def get_full_path_info(self, force_append_slash: bool = ...) -> str: ...
    def get_signed_cookie(
        self, key: str, default: Any = ..., salt: str = ..., max_age: int | datetime.timedelta | None = ...
    ) -> str | None: ...
    def build_absolute_uri(self, location: str | None = ...) -> str: ...
    @property
    def scheme(self) -> str | None: ...
    def is_secure(self) -> bool: ...
    @property
    def encoding(self) -> str | None: ...
    @encoding.setter
    def encoding(self, val: str) -> None: ...
    @property
    def upload_handlers(self) -> _UploadHandlerList: ...
    @upload_handlers.setter
    def upload_handlers(self, upload_handlers: _UploadHandlerList) -> None: ...
    @cached_property
    def accepted_types(self) -> list[MediaType]: ...
    def parse_file_upload(
        self, META: Mapping[str, Any], post_data: BinaryIO
    ) -> tuple[QueryDict, MultiValueDict[str, uploadedfile.UploadedFile]]: ...
    @cached_property
    def headers(self) -> HttpHeaders: ...
    @property
    def body(self) -> bytes: ...
    def _load_post_and_files(self) -> None: ...
    def accepts(self, media_type: str) -> bool: ...
    def readlines(self) -> list[bytes]: ...  # type: ignore[override]

@type_check_only
class _MutableHttpRequest(HttpRequest):
    GET: QueryDict  # type: ignore[assignment]
    POST: QueryDict  # type: ignore[assignment]

_Z = TypeVar("_Z")

class QueryDict(MultiValueDict[str, str]):
    _mutable: bool
    # We can make it mutable only by specifying `mutable=True`.
    # It can be done a) with kwarg and b) with pos. arg. `overload` has
    # some problems with args/kwargs + Literal, so two signatures are required.
    # ('querystring', True, [...])
    @overload
    def __init__(
        self: QueryDict,
        query_string: str | bytes | None,
        mutable: Literal[True],
        encoding: str | None = ...,
    ) -> None: ...
    # ([querystring='string',] mutable=True, [...])
    @overload
    def __init__(
        self: QueryDict,
        *,
        mutable: Literal[True],
        query_string: str | bytes | None = ...,
        encoding: str | None = ...,
    ) -> None: ...
    # Otherwise it's immutable
    @overload
    def __init__(  # type: ignore[misc]
        self: _ImmutableQueryDict,
        query_string: str | bytes | None = ...,
        mutable: bool = ...,
        encoding: str | None = ...,
    ) -> None: ...
    @classmethod
    def fromkeys(  # type: ignore[override]
        cls,
        iterable: Iterable[bytes | str],
        value: str | bytes = ...,
        mutable: bool = ...,
        encoding: str | None = ...,
    ) -> Self: ...
    @property
    def encoding(self) -> str: ...
    @encoding.setter
    def encoding(self, value: str) -> None: ...
    def __setitem__(self, key: str | bytes, value: str | bytes) -> None: ...
    def __delitem__(self, key: str | bytes) -> None: ...
    def setlist(self, key: str | bytes, list_: Iterable[str | bytes]) -> None: ...
    def setlistdefault(self, key: str | bytes, default_list: list[str] | None = ...) -> list[str]: ...
    def appendlist(self, key: str | bytes, value: str | bytes) -> None: ...
    # Fake signature (because *args is used in source, but it fails with more that 1 argument)
    @overload
    def pop(self, key: str | bytes, /) -> str: ...
    @overload
    def pop(self, key: str | bytes, default: str | _Z = ..., /) -> str | _Z: ...
    def popitem(self) -> tuple[str, str]: ...
    def clear(self) -> None: ...
    def setdefault(self, key: str | bytes, default: str | bytes | None = ...) -> str: ...
    def copy(self) -> QueryDict: ...
    def urlencode(self, safe: str | None = ...) -> str: ...

@type_check_only
class _ImmutableQueryDict(QueryDict):
    _mutable: Literal[False]
    # def __init__(
    #     self, query_string: Optional[Union[str, bytes]] = ..., mutable: bool = ..., encoding: Optional[str] = ...
    # ) -> None: ...
    def __setitem__(self, key: str | bytes, value: str | bytes) -> NoReturn: ...
    def __delitem__(self, key: str | bytes) -> NoReturn: ...
    def setlist(self, key: str | bytes, list_: Iterable[str | bytes]) -> NoReturn: ...
    def setlistdefault(self, key: str | bytes, default_list: list[str] | None = ...) -> NoReturn: ...
    def appendlist(self, key: str | bytes, value: str | bytes) -> NoReturn: ...
    # Fake signature (because *args is used in source, but it fails with more that 1 argument)
    @overload
    def pop(self, key: str | bytes, /) -> NoReturn: ...
    @overload
    def pop(self, key: str | bytes, default: str | _Z = ..., /) -> NoReturn: ...
    def popitem(self) -> NoReturn: ...
    def clear(self) -> NoReturn: ...
    def setdefault(self, key: str | bytes, default: str | bytes | None = ...) -> NoReturn: ...
    def copy(self) -> QueryDict: ...  # type: ignore[override]
    def urlencode(self, safe: str | None = ...) -> str: ...
    # Fakes for convenience (for `request.GET` and `request.POST`). If dict
    # was created by Django, there is no chance to hit `List[object]` (empty list)
    # edge case.
    def __getitem__(self, key: str) -> str: ...
    def dict(self) -> dict[str, str]: ...  # type: ignore[override]

class MediaType:
    main_type: str
    sub_type: str
    params: dict[str, bytes]
    def __init__(self, media_type_raw_line: str) -> None: ...
    @property
    def is_all_types(self) -> bool: ...
    def match(self, other: str) -> bool: ...

@overload
def bytes_to_text(s: None, encoding: str) -> None: ...
@overload
def bytes_to_text(s: bytes | str, encoding: str) -> str: ...
def split_domain_port(host: str) -> tuple[str, str]: ...
def validate_host(host: str, allowed_hosts: Iterable[str]) -> bool: ...
def parse_accept_header(header: str) -> list[MediaType]: ...



from collections import namedtuple
from typing import Any

from django.db.backends.base.introspection import BaseDatabaseIntrospection
from django.db.backends.mysql.base import DatabaseWrapper

FieldInfo: Any
InfoLine = namedtuple(
    "InfoLine",
    [
        "col_name",
        "data_type",
        "max_len",
        "num_prec",
        "num_scale",
        "extra",
        "column_default",
        "collation",
        "is_unsigned",
    ],
)

class DatabaseIntrospection(BaseDatabaseIntrospection):
    connection: DatabaseWrapper
    data_types_reverse: Any



__package__ = 'archivebox.config'

import sys
import shutil

from typing import Dict, Optional
from pathlib import Path

from rich import print
from pydantic import Field, field_validator, computed_field
from django.utils.crypto import get_random_string

from abx.archivebox.base_configset import BaseConfigSet


from .constants import CONSTANTS
from .version import get_COMMIT_HASH, get_BUILD_TIME
from .permissions import IN_DOCKER

###################### Config ##########################


class ShellConfig(BaseConfigSet):
    DEBUG: bool                         = Field(default=lambda: '--debug' in sys.argv)
    
    IS_TTY: bool                        = Field(default=sys.stdout.isatty())
    USE_COLOR: bool                     = Field(default=lambda c: c.IS_TTY)
    SHOW_PROGRESS: bool                 = Field(default=lambda c: c.IS_TTY)
    
    IN_DOCKER: bool                     = Field(default=IN_DOCKER)
    IN_QEMU: bool                       = Field(default=False)

    ANSI: Dict[str, str]                = Field(default=lambda c: CONSTANTS.DEFAULT_CLI_COLORS if c.USE_COLOR else CONSTANTS.DISABLED_CLI_COLORS)

    VERSIONS_AVAILABLE: bool = False             # .check_for_update.get_versions_available_on_github(c)},
    CAN_UPGRADE: bool = False                    # .check_for_update.can_upgrade(c)},

    
    @computed_field
    @property
    def TERM_WIDTH(self) -> int:
        if not self.IS_TTY:
            return 200
        return shutil.get_terminal_size((140, 10)).columns
    
    @computed_field
    @property
    def COMMIT_HASH(self) -> Optional[str]:
        return get_COMMIT_HASH()
    
    @computed_field
    @property
    def BUILD_TIME(self) -> str:
        return get_BUILD_TIME()

SHELL_CONFIG = ShellConfig()


class StorageConfig(BaseConfigSet):
    OUTPUT_PERMISSIONS: str             = Field(default='644')
    RESTRICT_FILE_NAMES: str            = Field(default='windows')
    ENFORCE_ATOMIC_WRITES: bool         = Field(default=True)
    
    # not supposed to be user settable:
    DIR_OUTPUT_PERMISSIONS: str         = Field(default=lambda c: c['OUTPUT_PERMISSIONS'].replace('6', '7').replace('4', '5'))


STORAGE_CONFIG = StorageConfig()


class GeneralConfig(BaseConfigSet):
    TAG_SEPARATOR_PATTERN: str          = Field(default=r'[,]')


GENERAL_CONFIG = GeneralConfig()


class ServerConfig(BaseConfigSet):
    SECRET_KEY: str                     = Field(default=lambda: get_random_string(50, 'abcdefghijklmnopqrstuvwxyz0123456789_'))
    BIND_ADDR: str                      = Field(default=lambda: ['127.0.0.1:8000', '0.0.0.0:8000'][SHELL_CONFIG.IN_DOCKER])
    ALLOWED_HOSTS: str                  = Field(default='*')
    CSRF_TRUSTED_ORIGINS: str           = Field(default=lambda c: 'http://localhost:8000,http://127.0.0.1:8000,http://0.0.0.0:8000,http://{}'.format(c.BIND_ADDR))
    
    SNAPSHOTS_PER_PAGE: int             = Field(default=40)
    FOOTER_INFO: str                    = Field(default='Content is hosted for personal archiving purposes only.  Contact server owner for any takedown requests.')
    # CUSTOM_TEMPLATES_DIR: Path          = Field(default=None)  # this is now a constant

    PUBLIC_INDEX: bool                  = Field(default=True)
    PUBLIC_SNAPSHOTS: bool              = Field(default=True)
    PUBLIC_ADD_VIEW: bool               = Field(default=False)
    
    ADMIN_USERNAME: str                 = Field(default=None)
    ADMIN_PASSWORD: str                 = Field(default=None)
    REVERSE_PROXY_USER_HEADER: str      = Field(default='Remote-User')
    REVERSE_PROXY_WHITELIST: str        = Field(default='')
    LOGOUT_REDIRECT_URL: str            = Field(default='/')
    PREVIEW_ORIGINALS: bool             = Field(default=True)
    
SERVER_CONFIG = ServerConfig()


class ArchivingConfig(BaseConfigSet):
    ONLY_NEW: bool                      = Field(default=True)
    
    TIMEOUT: int                        = Field(default=60)
    MEDIA_TIMEOUT: int                  = Field(default=3600)

    MEDIA_MAX_SIZE: str                 = Field(default='750m')
    RESOLUTION: str                     = Field(default='1440,2000')
    CHECK_SSL_VALIDITY: bool            = Field(default=True)
    USER_AGENT: str                     = Field(default='Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/128.0.0.0 Safari/537.36 ArchiveBox/{VERSION} (+https://github.com/ArchiveBox/ArchiveBox/)')
    COOKIES_FILE: Path | None           = Field(default=None)
    
    URL_DENYLIST: str                   = Field(default=r'\.(css|js|otf|ttf|woff|woff2|gstatic\.com|googleapis\.com/css)(\?.*)?$', alias='URL_BLACKLIST')
    URL_ALLOWLIST: str | None           = Field(default=None, alias='URL_WHITELIST')
    
    # GIT_DOMAINS: str                    = Field(default='github.com,bitbucket.org,gitlab.com,gist.github.com,codeberg.org,gitea.com,git.sr.ht')
    # WGET_USER_AGENT: str                = Field(default=lambda c: c['USER_AGENT'] + ' wget/{WGET_VERSION}')
    # CURL_USER_AGENT: str                = Field(default=lambda c: c['USER_AGENT'] + ' curl/{CURL_VERSION}')
    # CHROME_USER_AGENT: str              = Field(default=lambda c: c['USER_AGENT'])
    # CHROME_USER_DATA_DIR: str | None    = Field(default=None)
    # CHROME_TIMEOUT: int                 = Field(default=0)
    # CHROME_HEADLESS: bool               = Field(default=True)
    # CHROME_SANDBOX: bool                = Field(default=lambda: not SHELL_CONFIG.IN_DOCKER)

    @field_validator('TIMEOUT', mode='after')
    def validate_timeout(cls, v):
        if int(v) < 5:
            print(f'[red][!] Warning: TIMEOUT is set too low! (currently set to TIMEOUT={v} seconds)[/red]', file=sys.stderr)
            print('    You must allow *at least* 5 seconds for indexing and archive methods to run succesfully.', file=sys.stderr)
            print('    (Setting it to somewhere between 30 and 3000 seconds is recommended)', file=sys.stderr)
            print(file=sys.stderr)
            print('    If you want to make ArchiveBox run faster, disable specific archive methods instead:', file=sys.stderr)
            print('        https://github.com/ArchiveBox/ArchiveBox/wiki/Configuration#archive-method-toggles', file=sys.stderr)
            print(file=sys.stderr)
        return v
    
    @field_validator('CHECK_SSL_VALIDITY', mode='after')
    def validate_check_ssl_validity(cls, v):
        """SIDE EFFECT: disable "you really shouldnt disable ssl" warnings emitted by requests"""
        if not v:
            import requests
            import urllib3
            requests.packages.urllib3.disable_warnings(requests.packages.urllib3.exceptions.InsecureRequestWarning)
            urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
        return v

ARCHIVING_CONFIG = ArchivingConfig()


class SearchBackendConfig(BaseConfigSet):
    USE_INDEXING_BACKEND: bool          = Field(default=True)
    USE_SEARCHING_BACKEND: bool         = Field(default=True)
    
    SEARCH_BACKEND_ENGINE: str          = Field(default='ripgrep')
    SEARCH_PROCESS_HTML: bool           = Field(default=True)
    SEARCH_BACKEND_TIMEOUT: int         = Field(default=10)

SEARCH_BACKEND_CONFIG = SearchBackendConfig()




from collections.abc import Iterable, Mapping, Sequence
from datetime import datetime
from typing import Any, Generic, TypeVar

from django.contrib.sites.models import Site
from django.contrib.sites.requests import RequestSite
from django.core.paginator import Paginator
from django.db.models.base import Model
from django.db.models.query import QuerySet

_ItemT = TypeVar("_ItemT")

class Sitemap(Generic[_ItemT]):
    limit: int
    protocol: str | None
    i18n: bool
    languages: Sequence[str] | None
    alternates: bool
    x_default: bool
    def items(self) -> Iterable[_ItemT]: ...
    def location(self, item: _ItemT) -> str: ...
    @property
    def paginator(self) -> Paginator: ...
    def get_languages_for_item(self, item: _ItemT) -> list[str]: ...
    def get_protocol(self, protocol: str | None = ...) -> str: ...
    def get_domain(self, site: Site | RequestSite | None = ...) -> str: ...
    def get_urls(
        self, page: int | str = ..., site: Site | RequestSite | None = ..., protocol: str | None = ...
    ) -> list[dict[str, Any]]: ...
    def get_latest_lastmod(self) -> datetime | None: ...

_ModelT = TypeVar("_ModelT", bound=Model)

class GenericSitemap(Sitemap[_ModelT]):
    priority: float | None
    changefreq: str | None
    queryset: QuerySet[_ModelT]
    date_field: str | None
    protocol: str | None
    def __init__(
        self,
        info_dict: Mapping[str, datetime | QuerySet[_ModelT] | str],
        priority: float | None = ...,
        changefreq: str | None = ...,
        protocol: str | None = ...,
    ) -> None: ...
    def items(self) -> QuerySet[_ModelT]: ...
    def lastmod(self, item: _ModelT) -> datetime | None: ...
    def get_latest_lastmod(self) -> datetime | None: ...



from collections.abc import Iterable
from typing import Any

from django.db.backends.base.operations import BaseDatabaseOperations
from django.db.backends.postgresql.base import DatabaseWrapper

class DatabaseOperations(BaseDatabaseOperations):
    connection: DatabaseWrapper
    explain_options: frozenset[str]

    def bulk_insert_sql(self, fields: Any, placeholder_rows: Iterable[str]) -> str: ...
    def compose_sql(self, sql: Any, params: Any) -> Any: ...
    def fetch_returned_insert_rows(self, cursor: Any) -> Any: ...



from typing import Any, ClassVar

from django.core.cache.backends.base import BaseCache

class FileBasedCache(BaseCache):
    cache_suffix: ClassVar[str]
    pickle_protocol: ClassVar[int]
    def __init__(self, dir: str, params: dict[str, Any]) -> None: ...



from typing import Any

from django.db.backends.sqlite3.base import DatabaseWrapper as SQLiteDatabaseWrapper

from .client import SpatiaLiteClient
from .features import DatabaseFeatures
from .introspection import SpatiaLiteIntrospection
from .operations import SpatiaLiteOperations
from .schema import SpatialiteSchemaEditor

class DatabaseWrapper(SQLiteDatabaseWrapper):
    SchemaEditorClass: type[SpatialiteSchemaEditor]
    client_class: type[SpatiaLiteClient]
    features_class: type[DatabaseFeatures]
    introspection_class: type[SpatiaLiteIntrospection]
    ops_class: type[SpatiaLiteOperations]
    lib_spatialite_paths: str
    ops: SpatiaLiteOperations
    def __init__(self, *args: Any, **kwargs: Any) -> None: ...
    def get_new_connection(self, conn_params: Any) -> Any: ...
    def prepare_database(self) -> None: ...



__package__ = 'abx.archivebox'

import abx
import inspect
from pathlib import Path

from django.apps import AppConfig

from typing import List, Type, Dict
from typing_extensions import Self

from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    model_validator,
    InstanceOf,
    computed_field,
)
from benedict import benedict

from .base_hook import BaseHook, HookType

class BasePlugin(BaseModel):
    model_config = ConfigDict(
        extra='forbid',
        arbitrary_types_allowed=True,
        populate_by_name=True,
        from_attributes=True,
        validate_defaults=False,
        validate_assignment=False,
        revalidate_instances="always",
        # frozen=True,
    )

    # Required by AppConfig:
    app_label: str = Field()                      # e.g. 'singlefile'                  (one-word machine-readable representation, to use as url-safe id/db-table prefix_/attr name)
    verbose_name: str = Field()                   # e.g. 'SingleFile'                  (human-readable *short* label, for use in column names, form labels, etc.)
    docs_url: str = Field(default=None)           # e.g. 'https://github.com/...'
    
    # All the hooks the plugin will install:
    hooks: List[InstanceOf[BaseHook]] = Field(default=[])
    
    _is_registered: bool = False
    _is_ready: bool = False
    
    @computed_field
    @property
    def id(self) -> str:
        return self.__class__.__name__
    
    @property
    def name(self) -> str:
        return self.app_label
    
    # @computed_field
    @property
    def plugin_module(self) -> str:  # DottedImportPath
        """ "
        Dotted import path of the plugin's module (after its loaded via settings.INSTALLED_APPS).
        e.g. 'archivebox.plugins_pkg.npm.apps.NpmPlugin' -> 'plugins_pkg.npm'
        """
        return f"{self.__module__}.{self.__class__.__name__}".split("archivebox.", 1)[-1].rsplit('.apps.', 1)[0]


    @property
    def plugin_module_full(self) -> str:  # DottedImportPath
        """e.g. 'archivebox.plugins_pkg.npm.apps.NpmPlugin'"""
        return f"{self.__module__}.{self.__class__.__name__}"
    
    # @computed_field
    @property
    def plugin_dir(self) -> Path:
        return Path(inspect.getfile(self.__class__)).parent.resolve()
    
    @model_validator(mode='after')
    def validate(self) -> Self:
        """Validate the plugin's build-time configuration here before it's registered in Django at runtime."""
        
        # VERY IMPORTANT:
        # preserve references to original default objects,
        # pydantic deepcopies them by default which breaks mutability
        # see https://github.com/pydantic/pydantic/issues/7608
        # if we dont do this, then plugins_extractor.SINGLEFILE_CONFIG != settings.CONFIGS.SingleFileConfig for example
        # and calling .__init__() on one of them will not update the other
        self.hooks = self.model_fields['hooks'].default
        
        assert self.app_label and self.app_label and self.verbose_name, f'{self.__class__.__name__} is missing .name or .app_label or .verbose_name'
        
        # assert json.dumps(self.model_json_schema(), indent=4), f"Plugin {self.plugin_module} has invalid JSON schema."
        
        return self
    
    @property
    def AppConfig(plugin_self) -> Type[AppConfig]:
        """Generate a Django AppConfig class for this plugin."""


        class PluginAppConfig(AppConfig):
            """Django AppConfig for plugin, allows it to be loaded as a Django app listed in settings.INSTALLED_APPS."""
            name = plugin_self.plugin_module
            app_label = plugin_self.app_label
            verbose_name = plugin_self.verbose_name

            default_auto_field = 'django.db.models.AutoField'

            # handled by abx.hookimpl  ready()
            # def ready(self):
            #     from django.conf import settings
            #     plugin_self.ready(settings)

        return PluginAppConfig

    @property
    def HOOKS_BY_ID(self) -> Dict[str, InstanceOf[BaseHook]]:
        return benedict({hook.id: hook for hook in self.hooks})

    @property
    def HOOKS_BY_TYPE(self) -> Dict[HookType, Dict[str, InstanceOf[BaseHook]]]:
        hooks = benedict({})
        for hook in self.hooks:
            hooks[hook.hook_type] = hooks.get(hook.hook_type) or benedict({})
            hooks[hook.hook_type][hook.id] = hook
        return hooks



    @abx.hookimpl
    def register(self, settings):
        from archivebox.config.legacy import bump_startup_progress_bar

        self._is_registered = True
        bump_startup_progress_bar()

        # print('◣----------------- REGISTERED PLUGIN:', self.plugin_module, '-----------------◢')
        # print()

    @abx.hookimpl
    def ready(self, settings=None):
        """Runs any runtime code needed when AppConfig.ready() is called (after all models are imported)."""

        from archivebox.config.legacy import bump_startup_progress_bar

        assert self._is_registered, f"Tried to run {self.plugin_module}.ready() but it was never registered!"
        self._is_ready = True

        # settings.PLUGINS[self.id]._is_ready = True
        bump_startup_progress_bar()


    @abx.hookimpl
    def get_INSTALLED_APPS(self):
        return [self.plugin_module]





#!/usr/bin/env python3
__package__ = "pydantic_pkgr"

import os
import sys
import platform
from typing import Optional
from pathlib import Path

from pydantic import model_validator, TypeAdapter, computed_field

from .base_types import BinProviderName, PATHStr, BinName, InstallArgs, HostBinPath, bin_abspath
from .semver import SemVer
from .binprovider import BinProvider

OS = platform.system().lower()

DEFAULT_MACOS_DIR = Path('/opt/homebrew/bin') if platform.machine() == 'arm64' else Path('/usr/local/bin')
DEFAULT_LINUX_DIR = Path('/home/linuxbrew/.linuxbrew/bin')


class BrewProvider(BinProvider):
    name: BinProviderName = "brew"
    INSTALLER_BIN: BinName = "brew"
    
    PATH: PATHStr = "/home/linuxbrew/.linuxbrew/bin:/opt/homebrew/bin:/usr/local/bin"

    @model_validator(mode="after")
    def load_PATH(self):
        if not self.INSTALLER_BIN_ABSPATH:
            # brew is not availabe on this host
            self.PATH: PATHStr = ""
            return self

        PATHs = set()
        
        if OS == 'darwin' and os.path.isdir(DEFAULT_MACOS_DIR) and os.access(DEFAULT_MACOS_DIR, os.R_OK):
            PATHs.add(str(DEFAULT_MACOS_DIR))
        if OS != 'darwin' and os.path.isdir(DEFAULT_LINUX_DIR) and os.access(DEFAULT_LINUX_DIR, os.R_OK):
            PATHs.add(str(DEFAULT_LINUX_DIR))
        
        if not PATHs:
            # if we cant autodetect the paths, run brew --prefix to get the path manually (very slow)
            PATHs.add(self.exec(bin_name=self.INSTALLER_BIN_ABSPATH, cmd=["--prefix"]).stdout.strip() + "/bin")
        
        self.PATH = TypeAdapter(PATHStr).validate_python(':'.join(PATHs))
        return self

    def on_install(self, bin_name: str, packages: Optional[InstallArgs] = None, **context) -> str:
        packages = packages or self.on_get_packages(bin_name)

        if not self.INSTALLER_BIN_ABSPATH:
            raise Exception(f"{self.__class__.__name__}.INSTALLER_BIN is not available on this host: {self.INSTALLER_BIN}")

        # print(f'[*] {self.__class__.__name__}: Installing {bin_name}: {self.INSTALLER_BIN_ABSPATH} install {packages}')

        # Attempt 1: Try installing with Pyinfra
        from .binprovider_pyinfra import PYINFRA_INSTALLED, pyinfra_package_install

        if PYINFRA_INSTALLED:
            return pyinfra_package_install(bin_name, installer_module="operations.brew.packages")

        # Attempt 2: Try installing with Ansible
        from .binprovider_ansible import ANSIBLE_INSTALLED, ansible_package_install

        if ANSIBLE_INSTALLED:
            return ansible_package_install(bin_name, installer_module="community.general.homebrew")

        # Attempt 3: Fallback to installing manually by calling brew in shell
        self.exec(bin_name=self.INSTALLER_BIN_ABSPATH, cmd=["update"])
        proc = self.exec(bin_name=self.INSTALLER_BIN_ABSPATH, cmd=["install", *packages])
        if proc.returncode != 0:
            print(proc.stdout.strip())
            print(proc.stderr.strip())
            raise Exception(f"{self.__class__.__name__} install got returncode {proc.returncode} while installing {packages}: {packages}")

        return proc.stderr.strip() + "\n" + proc.stdout.strip()

    def on_get_abspath(self, bin_name: BinName | HostBinPath, **context) -> HostBinPath | None:
        # print(f'[*] {self.__class__.__name__}: Getting abspath for {bin_name}...')

        if not self.PATH:
            return None
        
        # not all brew-installed binaries are symlinked into the default bin dir (e.g. curl)
        # because it might conflict with a system binary of the same name (e.g. /usr/bin/curl)
        # so we need to check for the binary in the namespaced opt dir as well
        extra_path = self.PATH.replace('/bin', '/opt/{bin_name}/bin')     # e.g. /opt/homebrew/opt/curl/bin/curl
        
        abspath = bin_abspath(bin_name, PATH=f'{self.PATH}:{extra_path}')
        if abspath:
            return abspath
        
        if not self.INSTALLER_BIN_ABSPATH:
            return None
        
        # fallback to using brew info to get the Cellar bin path
        for package in (self.on_get_packages(str(bin_name)) or [str(bin_name)]):
            try:
                info_lines = self.exec(bin_name=self.INSTALLER_BIN_ABSPATH, cmd=['info', '--quiet', package], timeout=5, quiet=True).stdout.strip().split('\n')
                # /opt/homebrew/Cellar/curl/8.10.0 (530 files, 4MB)
                cellar_path = [line for line in info_lines if '/Cellar/' in line][0].rsplit(' (', 1)[0]
                abspath = bin_abspath(bin_name, PATH=f'{cellar_path}/bin')
                if abspath:
                    return abspath
            except Exception:
                pass
        return None
        

    def on_get_version(self, bin_name: BinName, abspath: Optional[HostBinPath]=None, **context) -> SemVer | None:
        # print(f'[*] {self.__class__.__name__}: Getting version for {bin_name}...')
        try:
            version =  super().on_get_version(bin_name, abspath, **context)
            if version:
                return version
        except ValueError:
            pass
        
        if not self.INSTALLER_BIN_ABSPATH:
            return None
        
        # fallback to using brew info to get the version
        package = (self.on_get_packages(str(bin_name)) or [str(bin_name)])[-1]   # assume last package in list is the main one
        try:
            version_str = self.exec(bin_name=self.INSTALLER_BIN_ABSPATH, cmd=['info', '--quiet', package], quiet=True, timeout=5).stdout.strip()
            return SemVer.parse(version_str)
        except Exception:
            return None

if __name__ == "__main__":
    # Usage:
    # ./binprovider_brew.py load yt-dlp
    # ./binprovider_brew.py install pip
    # ./binprovider_brew.py get_version pip
    # ./binprovider_brew.py get_abspath pip
    result = brew = BrewProvider()

    if len(sys.argv) > 1:
        result = func = getattr(brew, sys.argv[1])  # e.g. install

    if len(sys.argv) > 2:
        result = func(sys.argv[2])  # e.g. install ffmpeg

    print(result)



# Generated by Django 3.1.3 on 2021-04-01 06:54

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0018_auto_20210327_0952'),
    ]

    operations = [
        migrations.AlterField(
            model_name='snapshot',
            name='url',
            field=models.URLField(db_index=True, unique=True),
        ),
    ]



from collections.abc import Callable
from typing import TypeVar

from django.contrib.sitemaps import Sitemap
from django.http.request import HttpRequest
from django.template.response import TemplateResponse

_C = TypeVar("_C", bound=Callable)

def x_robots_tag(func: _C) -> _C: ...
def index(
    request: HttpRequest,
    sitemaps: dict[str, type[Sitemap] | Sitemap],
    template_name: str = ...,
    content_type: str = ...,
    sitemap_url_name: str = ...,
) -> TemplateResponse: ...
def sitemap(
    request: HttpRequest,
    sitemaps: dict[str, type[Sitemap] | Sitemap],
    section: str | None = ...,
    template_name: str = ...,
    content_type: str = ...,
) -> TemplateResponse: ...



import sys
from _ctypes import _CData
from collections.abc import Callable, Sequence
from ctypes import CDLL, _FuncPointer, _NamedFuncPointer
from logging import Logger
from typing import Any

logger: Logger
lib_path: str | None
lib_names: list[str] | None
lib_name: str
lgdal: CDLL
if sys.platform == "win32":
    from ctypes import WinDLL

    lwingdal: WinDLL

def std_call(func: str) -> _NamedFuncPointer: ...
def gdal_version() -> bytes: ...
def gdal_full_version() -> bytes: ...
def gdal_version_info() -> tuple[int, int, int]: ...

GDAL_VERSION: tuple[int, int, int]
CPLErrorHandler: type[_FuncPointer]

def function(
    name: str,
    # Taken from _ctypes.CFuncPtr
    args: Sequence[type[_CData]],
    restype: type[_CData] | Callable[[int], Any] | None,
) -> _NamedFuncPointer: ...

err_handler: _FuncPointer
set_error_handler: _NamedFuncPointer



__package__ = 'archivebox.plugins_search.sonic'

import sys
from typing import List, Dict, Generator, cast

# Depends on other PyPI/vendor packages:
from pydantic import InstanceOf, Field, model_validator
from pydantic_pkgr import BinProvider, BinProviderName, ProviderLookupDict, BinName

# Depends on other Django apps:
from abx.archivebox.base_plugin import BasePlugin
from abx.archivebox.base_configset import BaseConfigSet
from abx.archivebox.base_binary import BaseBinary, env, brew
from abx.archivebox.base_hook import BaseHook
from abx.archivebox.base_searchbackend import BaseSearchBackend

# Depends on Other Plugins:
from archivebox.config.common import SEARCH_BACKEND_CONFIG

SONIC_LIB = None
try:
    import sonic
    SONIC_LIB = sonic
except ImportError:
    SONIC_LIB = None

###################### Config ##########################

class SonicConfig(BaseConfigSet):
    SONIC_BINARY: str       = Field(default='sonic')
    
    SONIC_HOST: str         = Field(default='localhost', alias='SEARCH_BACKEND_HOST_NAME')
    SONIC_PORT: int         = Field(default=1491, alias='SEARCH_BACKEND_PORT')
    SONIC_PASSWORD: str     = Field(default='SecretPassword', alias='SEARCH_BACKEND_PASSWORD')
    SONIC_COLLECTION: str   = Field(default='archivebox')
    SONIC_BUCKET: str       = Field(default='archivebox')
    
    SONIC_MAX_CHUNK_LENGTH: int     = Field(default=2000)
    SONIC_MAX_TEXT_LENGTH: int      = Field(default=100000000)
    SONIC_MAX_RETRIES: int          = Field(default=5)

    @model_validator(mode='after')
    def validate_sonic_port(self):
        if SEARCH_BACKEND_CONFIG.SEARCH_BACKEND_ENGINE == 'sonic' and SONIC_LIB is None:
            sys.stderr.write('[X] Error: Sonic search backend is enabled but sonic-client lib is not installed. You may need to run: pip install archivebox[sonic]\n')
            # dont hard exit here. in case the user is just running "archivebox version" or "archivebox help", we still want those to work despite broken ldap
            # sys.exit(1)
            SEARCH_BACKEND_CONFIG.update_in_place(SEARCH_BACKEND_ENGINE='ripgrep')
        return self

SONIC_CONFIG = SonicConfig()



class SonicBinary(BaseBinary):
    name: BinName = SONIC_CONFIG.SONIC_BINARY
    binproviders_supported: List[InstanceOf[BinProvider]] = [brew, env]   # TODO: add cargo

    provider_overrides: Dict[BinProviderName, ProviderLookupDict] = {
        brew.name: {'packages': lambda: ['sonic']},
        # cargo.name: {'packages': lambda: ['sonic-server']},             # TODO: add cargo
    }
    
    # TODO: add version checking over protocol? for when sonic backend is on remote server and binary is not installed locally
    # def on_get_version(self):
    #     with sonic.IngestClient(SONIC_CONFIG.SONIC_HOST, str(SONIC_CONFIG.SONIC_PORT), SONIC_CONFIG.SONIC_PASSWORD) as ingestcl:
    #         return SemVer.parse(str(ingestcl.protocol))

SONIC_BINARY = SonicBinary()



class SonicSearchBackend(BaseSearchBackend):
    name: str = 'sonic'
    docs_url: str = 'https://github.com/valeriansaliou/sonic'
    
    @staticmethod
    def index(snapshot_id: str, texts: List[str]):
        error_count = 0
        with sonic.IngestClient(SONIC_CONFIG.SONIC_HOST, str(SONIC_CONFIG.SONIC_PORT), SONIC_CONFIG.SONIC_PASSWORD) as ingestcl:
            for text in texts:
                chunks = (
                    text[i:i+SONIC_CONFIG.SONIC_MAX_CHUNK_LENGTH]
                    for i in range(
                        0,
                        min(len(text), SONIC_CONFIG.SONIC_MAX_TEXT_LENGTH),
                        SONIC_CONFIG.SONIC_MAX_CHUNK_LENGTH,
                    )
                )
                try:
                    for chunk in chunks:
                        ingestcl.push(SONIC_CONFIG.SONIC_COLLECTION, SONIC_CONFIG.SONIC_BUCKET, snapshot_id, str(chunk))
                except Exception as err:
                    print(f'[!] Sonic search backend threw an error while indexing: {err.__class__.__name__} {err}')
                    error_count += 1
                    if error_count > SONIC_CONFIG.SONIC_MAX_RETRIES:
                        raise

    @staticmethod
    def flush(snapshot_ids: Generator[str, None, None]):
        with sonic.IngestClient(SONIC_CONFIG.SONIC_HOST, str(SONIC_CONFIG.SONIC_PORT), SONIC_CONFIG.SONIC_PASSWORD) as ingestcl:
            for id in snapshot_ids:
                ingestcl.flush_object(SONIC_CONFIG.SONIC_COLLECTION, SONIC_CONFIG.SONIC_BUCKET, str(id))
    

    @staticmethod
    def search(text: str) -> List[str]:
        with sonic.SearchClient(SONIC_CONFIG.SONIC_HOST, SONIC_CONFIG.SONIC_PORT, SONIC_CONFIG.SONIC_PASSWORD) as querycl:
            snap_ids = cast(List[str], querycl.query(SONIC_CONFIG.SONIC_COLLECTION, SONIC_CONFIG.SONIC_BUCKET, text))
        return [str(id) for id in snap_ids]
    
    
SONIC_SEARCH_BACKEND = SonicSearchBackend()




class SonicSearchPlugin(BasePlugin):
    app_label: str ='sonic'
    verbose_name: str = 'Sonic'

    hooks: List[InstanceOf[BaseHook]] = [
        SONIC_CONFIG,
        SONIC_BINARY,
        SONIC_SEARCH_BACKEND,
    ]



PLUGIN = SonicSearchPlugin()
# PLUGIN.register(settings)
DJANGO_APP = PLUGIN.AppConfig



__package__ = 'pydantic_pkgr'


# Unfortunately it must be kept in the same file as BinProvider because of the circular type reference between them
from .binprovider import ShallowBinary      # noqa: F401



from django.core.exceptions import SuspiciousOperation

class DisallowedModelAdminLookup(SuspiciousOperation): ...
class DisallowedModelAdminToField(SuspiciousOperation): ...
class AlreadyRegistered(Exception): ...
class NotRegistered(Exception): ...



__package__ = 'abx.archivebox'

import os
from pathlib import Path
from typing import Type, Tuple, Callable, ClassVar

from benedict import benedict
from pydantic import model_validator, TypeAdapter
from pydantic_settings import BaseSettings, SettingsConfigDict, PydanticBaseSettingsSource
from pydantic_settings.sources import TomlConfigSettingsSource

from pydantic_pkgr.base_types import func_takes_args_or_kwargs

import abx

from .base_hook import BaseHook, HookType
from . import toml_util


PACKAGE_DIR = Path(__file__).resolve().parent.parent
DATA_DIR = Path(os.getcwd()).resolve()



class FlatTomlConfigSettingsSource(TomlConfigSettingsSource):
    """
    A source class that loads variables from a TOML file
    """

    def __init__(
        self,
        settings_cls: type[BaseSettings],
        toml_file: Path | None=None,
    ):
        self.toml_file_path = toml_file or settings_cls.model_config.get("toml_file")
        
        self.nested_toml_data = self._read_files(self.toml_file_path)
        self.toml_data = {}
        for top_level_key, top_level_value in self.nested_toml_data.items():
            if isinstance(top_level_value, dict):
                # value is nested, flatten it
                for key, value in top_level_value.items():
                    self.toml_data[key] = value
            else:
                # value is already flat, just set it as-is
                self.toml_data[top_level_key] = top_level_value
                
        # filter toml_data to only include keys that are defined on this settings_cls
        self.toml_data = {
            key: value
            for key, value in self.toml_data.items()
            if key in settings_cls.model_fields
        }
            
        super(TomlConfigSettingsSource, self).__init__(settings_cls, self.toml_data)


class ArchiveBoxBaseConfig(BaseSettings):
    """
    This is the base class for an ArchiveBox ConfigSet.
    It handles loading values from schema defaults, ArchiveBox.conf TOML config, and environment variables.

    class WgetConfig(ArchiveBoxBaseConfig):
        WGET_BINARY: str = Field(default='wget', alias='WGET_BINARY_PATH')

    c = WgetConfig()
    print(c.WGET_BINARY)                    # outputs: wget

    # you can mutate process environment variable and reload config using .__init__()
    os.environ['WGET_BINARY_PATH'] = 'wget2'
    c.__init__()

    print(c.WGET_BINARY)                    # outputs: wget2

    """
    
    # these pydantic config options are all VERY carefully chosen, make sure to test thoroughly before changing!!!
    model_config = SettingsConfigDict(
        validate_default=False,
        case_sensitive=True,
        extra="ignore",
        arbitrary_types_allowed=False,
        populate_by_name=True,
        from_attributes=True,
        loc_by_alias=False,
        validate_assignment=True,
        validate_return=True,
        revalidate_instances="always",
    )
    
    load_from_defaults: ClassVar[bool] = True
    load_from_configfile: ClassVar[bool] = True
    load_from_environment: ClassVar[bool] = True

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: Type[BaseSettings],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ) -> Tuple[PydanticBaseSettingsSource, ...]:
        """Defines the config precedence order: Schema defaults -> ArchiveBox.conf (TOML) -> Environment variables"""
        
        ARCHIVEBOX_CONFIG_FILE = DATA_DIR / "ArchiveBox.conf"
        ARCHIVEBOX_CONFIG_FILE_BAK = ARCHIVEBOX_CONFIG_FILE.parent / ".ArchiveBox.conf.bak"
        
        # import ipdb; ipdb.set_trace()
        
        precedence_order = {}
        
        # if ArchiveBox.conf does not exist yet, return defaults -> env order
        if not ARCHIVEBOX_CONFIG_FILE.is_file():
            precedence_order = {
                'defaults': init_settings,
                'environment': env_settings,
            }
        
        # if ArchiveBox.conf exists and is in TOML format, return default -> TOML -> env order
        try:
            precedence_order = precedence_order or {
                'defaults': init_settings,
                'configfile': FlatTomlConfigSettingsSource(settings_cls, toml_file=ARCHIVEBOX_CONFIG_FILE),
                'environment': env_settings,
            }
        except Exception as err:
            if err.__class__.__name__ != "TOMLDecodeError":
                raise
            # if ArchiveBox.conf exists and is in INI format, convert it then return default -> TOML -> env order

            # Convert ArchiveBox.conf in INI format to TOML and save original to .ArchiveBox.bak
            original_ini = ARCHIVEBOX_CONFIG_FILE.read_text()
            ARCHIVEBOX_CONFIG_FILE_BAK.write_text(original_ini)
            new_toml = toml_util.convert(original_ini)
            ARCHIVEBOX_CONFIG_FILE.write_text(new_toml)

            precedence_order = {
                'defaults': init_settings,
                'configfile': FlatTomlConfigSettingsSource(settings_cls, toml_file=ARCHIVEBOX_CONFIG_FILE),
                'environment': env_settings,
            }
            
        if not cls.load_from_environment:
            precedence_order.pop('environment')
        if not cls.load_from_configfile:
            precedence_order.pop('configfile')
        if not cls.load_from_defaults:
            precedence_order.pop('defaults')

        return tuple(precedence_order.values())

    @model_validator(mode="after")
    def fill_defaults(self):
        """Populate any unset values using function provided as their default"""

        for key, field in self.model_fields.items():
            value = getattr(self, key)
            
            if isinstance(value, Callable):
                # if value is a function, execute it to get the actual value, passing existing config as a dict arg if expected
                if func_takes_args_or_kwargs(value):
                    # assemble dict of existing field values to pass to default factory functions
                    config_so_far = benedict(self.model_dump(include=set(self.model_fields.keys()), warnings=False))
                    computed_default = field.default(config_so_far)
                else:
                    # otherwise it's a pure function with no args, just call it
                    computed_default = field.default()

                # coerce/check to make sure default factory return value matches type annotation
                TypeAdapter(field.annotation).validate_python(computed_default)

                # set generated default value as final validated value
                setattr(self, key, computed_default)
        return self
    
    def update_in_place(self, warn=True, **kwargs):
        """
        Update the config with new values. Use this sparingly! We should almost never be updating config at runtime.
        Sets them in the environment so they propagate to spawned subprocesses / across future re-__init__()s and reload from environment

        Example acceptable use case: user config says SEARCH_BACKEND_ENGINE=sonic but sonic_client pip library is not installed so we cannot use it.
        SEARCH_BACKEND_CONFIG.update_in_place(SEARCH_BACKEND_ENGINE='ripgrep') can be used to reset it back to ripgrep so we can continue.
        """
        if warn:
            print('[!] WARNING: Some of the provided user config values cannot be used, temporarily ignoring them:')
        for key, value in kwargs.items():
            os.environ[key] = str(value)
            original_value = getattr(self, key)
            if warn:
                print(f'    {key}={original_value} -> {value}')
        self.__init__()
        return self
    
    def as_legacy_config_schema(self):
        # shim for backwards compatibility with old config schema style
        model_values = self.model_dump()
        return benedict({
            key: {'type': field.annotation, 'default': model_values[key]}
            for key, field in self.model_fields.items()
        })


class BaseConfigSet(ArchiveBoxBaseConfig, BaseHook):      # type: ignore[type-arg]
    hook_type: ClassVar[HookType] = 'CONFIG'

    # @abx.hookimpl
    # def ready(self, settings):
    #    # reload config from environment, in case it's been changed by any other plugins
    #    self.__init__()


    @abx.hookimpl
    def get_CONFIGS(self):
        try:
            return {self.id: self}
        except Exception as e:
            # raise Exception(f'Error computing CONFIGS for {type(self)}: {e.__class__.__name__}: {e}')
            print(f'Error computing CONFIGS for {type(self)}: {e.__class__.__name__}: {e}')
        return {}

    @abx.hookimpl
    def get_FLAT_CONFIG(self):
        try:
            return self.model_dump()
        except Exception as e:
            # raise Exception(f'Error computing FLAT_CONFIG for {type(self)}: {e.__class__.__name__}: {e}')
            print(f'Error computing FLAT_CONFIG for {type(self)}: {e.__class__.__name__}: {e}')
        return {}



__package__ = 'archivebox.parsers'


import re

from typing import IO, Iterable, Optional
from datetime import datetime, timezone

from ..index.schema import Link
from archivebox.misc.util import (
    htmldecode,
    enforce_types,
    find_all_urls,
)
from html.parser import HTMLParser
from urllib.parse import urljoin


class HrefParser(HTMLParser):
    def __init__(self):
        super().__init__()
        self.urls = []

    def handle_starttag(self, tag, attrs):
        if tag == "a":
            for attr, value in attrs:
                if attr == "href":
                    self.urls.append(value)


@enforce_types
def parse_generic_html_export(html_file: IO[str], root_url: Optional[str]=None, **_kwargs) -> Iterable[Link]:
    """Parse Generic HTML for href tags and use only the url (support for title coming later)"""

    html_file.seek(0)
    for line in html_file:
        parser = HrefParser()
        # example line
        # <li><a href="http://example.com/ time_added="1478739709" tags="tag1,tag2">example title</a></li>
        parser.feed(line)
        for url in parser.urls:
            if root_url:
                url_is_absolute = (url.lower().startswith('http://') or url.lower().startswith('https://'))
                # url = https://abc.com                       => True
                # url = /page.php?next=https://example.com    => False

                if not url_is_absolute:                       # resolve it by joining it with root_url
                    relative_path = url

                    url = urljoin(root_url, relative_path)    # https://example.com/somepage.html + /home.html
                                                              # => https://example.com/home.html

                    # special case to handle bug around // handling, crucial for urls that contain sub-urls
                    # e.g. https://web.archive.org/web/https://example.com
                    if did_urljoin_misbehave(root_url, relative_path, url):
                        url = fix_urljoin_bug(url)

            for archivable_url in find_all_urls(url):
                yield Link(
                    url=htmldecode(archivable_url),
                    timestamp=str(datetime.now(timezone.utc).timestamp()),
                    title=None,
                    tags=None,
                    sources=[html_file.name],
                )


KEY = 'html'
NAME = 'Generic HTML'
PARSER = parse_generic_html_export


#### WORKAROUND CODE FOR https://github.com/python/cpython/issues/96015 ####

def did_urljoin_misbehave(root_url: str, relative_path: str, final_url: str) -> bool:
    """
    Handle urljoin edge case bug where multiple slashes get turned into a single slash:
    - https://github.com/python/cpython/issues/96015
    - https://github.com/ArchiveBox/ArchiveBox/issues/1411

    This workaround only fixes the most common case of a sub-URL inside an outer URL, e.g.:
       https://web.archive.org/web/https://example.com/some/inner/url

    But there are other valid URLs containing // that are not fixed by this workaround, e.g.:
       https://example.com/drives/C//some/file
    """

    # if relative path is actually an absolute url, cut off its own scheme so we check the path component only
    relative_path = relative_path.lower()
    if relative_path.startswith('http://') or relative_path.startswith('https://'):
        relative_path = relative_path.split('://', 1)[-1]

    # TODO: properly fix all double // getting stripped by urljoin, not just ://
    original_path_had_suburl = '://' in relative_path
    original_root_had_suburl = '://' in root_url[8:]     # ignore first 8 chars because root always starts with https://
    final_joined_has_suburl = '://' in final_url[8:]     # ignore first 8 chars because final always starts with https://

    urljoin_broke_suburls = (
        (original_root_had_suburl or original_path_had_suburl)
        and not final_joined_has_suburl
    )
    return urljoin_broke_suburls


def fix_urljoin_bug(url: str, nesting_limit=5):
    """
    recursively replace broken suburls .../http:/... with .../http://...

    basically equivalent to this for 99.9% of cases:
      url = url.replace('/http:/',  '/http://')
      url = url.replace('/https:/', '/https://')
    except this handles:
        other schemes besides http/https     (e.g. https://example.com/link/git+ssh://github.com/example)
        other preceding separators besides / (e.g. https://example.com/login/?next=https://example.com/home)
        fixing multiple suburls recursively
    """
    input_url = url
    for _ in range(nesting_limit):
        url = re.sub(
            r'(?P<root>.+?)'                             # https://web.archive.org/web
            + r'(?P<separator>[-=/_&+%$#@!*\(\\])'       # /
            + r'(?P<subscheme>[a-zA-Z0-9+_-]{1,32}?):/'  # http:/
            + r'(?P<suburl>[^/\\]+)',                    # example.com
            r"\1\2\3://\4",
            input_url,
            re.IGNORECASE | re.UNICODE,
        )
        if url == input_url:
            break                                        # nothing left to replace, all suburls are fixed
        input_url = url

    return url


# sanity check to make sure workaround code works as expected and doesnt introduce *more* bugs
assert did_urljoin_misbehave('https://web.archive.org/web/https://example.com', 'abc.html', 'https://web.archive.org/web/https:/example.com/abc.html') == True
assert did_urljoin_misbehave('http://example.com', 'https://web.archive.org/web/http://example.com/abc.html', 'https://web.archive.org/web/http:/example.com/abc.html') == True
assert fix_urljoin_bug('https:/example.com') == 'https:/example.com'   # should not modify original url's scheme, only sub-urls
assert fix_urljoin_bug('https://web.archive.org/web/https:/example.com/abc.html') == 'https://web.archive.org/web/https://example.com/abc.html'
assert fix_urljoin_bug('http://example.com/link/git+ssh:/github.com/example?next=ftp:/example.com') == 'http://example.com/link/git+ssh://github.com/example?next=ftp://example.com'




from collections import namedtuple
from collections.abc import Iterator
from typing import Literal

from django.db.migrations.state import ModelState, ProjectState
from django.db.models import Field, Model

def resolve_relation(
    model: str | type[Model], app_label: str | None = ..., model_name: str | None = ...
) -> tuple[str, str]: ...

FieldReference = namedtuple("FieldReference", ["to", "through"])

def field_references(
    model_tuple: tuple[str, str],
    field: Field,
    reference_model_tuple: tuple[str, str],
    reference_field_name: str | None = ...,
    reference_field: Field | None = ...,
) -> Literal[False] | FieldReference: ...
def get_references(
    state: ProjectState,
    model_tuple: tuple[str, str],
    field_tuple: tuple[()] | tuple[str, Field] = ...,
) -> Iterator[tuple[ModelState, str, Field, FieldReference]]: ...
def field_is_referenced(state: ProjectState, model_tuple: tuple[str, str], field_tuple: tuple[str, Field]) -> bool: ...



from typing import Any

from django.db.backends.postgresql.schema import DatabaseSchemaEditor

class PostGISSchemaEditor(DatabaseSchemaEditor):
    geom_index_type: str
    geom_index_ops_nd: str
    rast_index_wrapper: str
    sql_alter_column_to_3d: str
    sql_alter_column_to_2d: str
    def geo_quote_name(self, name: Any) -> Any: ...



__package__ = 'archivebox.api'

# from django.test import TestCase
# from ninja.testing import TestClient

# from .routes_cli import router

# class ArchiveBoxCLIAPITestCase(TestCase):
#     def setUp(self):
#         self.client = TestClient(router)

#     def test_add_endpoint(self):
#         response = self.client.post("/add", json={"urls": ["http://example.com"], "tag": "testTag1,testTag2"})
#         self.assertEqual(response.status_code, 200)
#         self.assertTrue(response.json()["success"])

#     def test_remove_endpoint(self):
#         response = self.client.post("/remove", json={"filter_patterns": ["http://example.com"]})
#         self.assertEqual(response.status_code, 200)
#         self.assertTrue(response.json()["success"])

#     def test_update_endpoint(self):
#         response = self.client.post("/update", json={})
#         self.assertEqual(response.status_code, 200)
#         self.assertTrue(response.json()["success"])

#     def test_list_all_endpoint(self):
#         response = self.client.post("/list_all", json={})
#         self.assertEqual(response.status_code, 200)
#         self.assertTrue(response.json()["success"])



from typing import Any, ClassVar

from django.db.models import Aggregate, FloatField, IntegerField

class StatAggregate(Aggregate):
    output_field: ClassVar[FloatField]
    def __init__(
        self, y: Any, x: Any, output_field: Any | None = ..., filter: Any | None = ..., default: Any | None = ...
    ) -> None: ...

class Corr(StatAggregate): ...

class CovarPop(StatAggregate):
    def __init__(
        self, y: Any, x: Any, sample: bool = ..., filter: Any | None = ..., default: Any | None = ...
    ) -> None: ...

class RegrAvgX(StatAggregate): ...
class RegrAvgY(StatAggregate): ...

class RegrCount(StatAggregate):
    output_field: ClassVar[IntegerField]  # type: ignore[assignment]

class RegrIntercept(StatAggregate): ...
class RegrR2(StatAggregate): ...
class RegrSlope(StatAggregate): ...
class RegrSXX(StatAggregate): ...
class RegrSXY(StatAggregate): ...
class RegrSYY(StatAggregate): ...



# Generated by Django 5.0.6 on 2024-08-18 05:34

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0032_alter_archiveresult_id'),
    ]

    operations = [
        migrations.RenameField(
            model_name='archiveresult',
            old_name='id',
            new_name='old_id',
        ),
    ]



__package__ = 'archivebox.index'

from pathlib import Path
from datetime import datetime, timezone
from collections import defaultdict
from typing import List, Optional, Iterator, Mapping

from django.utils.html import format_html, mark_safe   # type: ignore
from django.core.cache import cache

from archivebox.misc.system import atomic_write
from archivebox.misc.util import (
    enforce_types,
    ts_to_date_str,
    urlencode,
    htmlencode,
    urldecode,
)
from archivebox.config import CONSTANTS, DATA_DIR, VERSION
from archivebox.config.common import SERVER_CONFIG
from archivebox.config.version import get_COMMIT_HASH
from archivebox.plugins_extractor.archivedotorg.apps import ARCHIVEDOTORG_CONFIG

from .schema import Link
from ..logging_util import printable_filesize

MAIN_INDEX_TEMPLATE = 'static_index.html'
MINIMAL_INDEX_TEMPLATE = 'minimal_index.html'
LINK_DETAILS_TEMPLATE = 'snapshot.html'
TITLE_LOADING_MSG = 'Not yet archived...'


### Main Links Index

@enforce_types
def parse_html_main_index(out_dir: Path=DATA_DIR) -> Iterator[str]:
    """parse an archive index html file and return the list of urls"""

    index_path = Path(out_dir) / CONSTANTS.HTML_INDEX_FILENAME
    if index_path.exists():
        with open(index_path, 'r', encoding='utf-8') as f:
            for line in f:
                if 'class="link-url"' in line:
                    yield line.split('"')[1]
    return ()

@enforce_types
def generate_index_from_links(links: List[Link], with_headers: bool):
    if with_headers:
        output = main_index_template(links)
    else:
        output = main_index_template(links, template=MINIMAL_INDEX_TEMPLATE)
    return output

@enforce_types
def main_index_template(links: List[Link], template: str=MAIN_INDEX_TEMPLATE) -> str:
    """render the template for the entire main index"""

    return render_django_template(template, {
        'version': VERSION,
        'git_sha': get_COMMIT_HASH() or VERSION,
        'num_links': str(len(links)),
        'date_updated': datetime.now(timezone.utc).strftime('%Y-%m-%d'),
        'time_updated': datetime.now(timezone.utc).strftime('%Y-%m-%d %H:%M'),
        'links': [link._asdict(extended=True) for link in links],
        'FOOTER_INFO': SERVER_CONFIG.FOOTER_INFO,
    })


### Link Details Index

@enforce_types
def write_html_link_details(link: Link, out_dir: Optional[str]=None) -> None:
    out_dir = out_dir or link.link_dir

    rendered_html = link_details_template(link)
    atomic_write(str(Path(out_dir) / CONSTANTS.HTML_INDEX_FILENAME), rendered_html)


@enforce_types
def link_details_template(link: Link) -> str:

    from ..extractors.wget import wget_output_path

    link_info = link._asdict(extended=True)

    return render_django_template(LINK_DETAILS_TEMPLATE, {
        **link_info,
        **link_info['canonical'],
        'title': htmlencode(
            link.title
            or (link.base_url if link.is_archived else TITLE_LOADING_MSG)
        ),
        'url_str': htmlencode(urldecode(link.base_url)),
        'archive_url': urlencode(
            wget_output_path(link)
            or (link.domain if link.is_archived else '')
        ) or 'about:blank',
        'extension': link.extension or 'html',
        'tags': link.tags or 'untagged',
        'size': printable_filesize(link.archive_size) if link.archive_size else 'pending',
        'status': 'archived' if link.is_archived else 'not yet archived',
        'status_color': 'success' if link.is_archived else 'danger',
        'oldest_archive_date': ts_to_date_str(link.oldest_archive_date),
        'SAVE_ARCHIVE_DOT_ORG': ARCHIVEDOTORG_CONFIG.SAVE_ARCHIVE_DOT_ORG,
        'PREVIEW_ORIGINALS': SERVER_CONFIG.PREVIEW_ORIGINALS,
    })

@enforce_types
def render_django_template(template: str, context: Mapping[str, str]) -> str:
    """render a given html template string with the given template content"""
    from django.template.loader import render_to_string

    return render_to_string(template, context)


def snapshot_icons(snapshot) -> str:
    cache_key = f'result_icons:{snapshot.pk}:{(snapshot.downloaded_at or snapshot.modified_at or snapshot.created_at or snapshot.bookmarked_at).timestamp()}'
    
    def calc_snapshot_icons():
        from core.models import ArchiveResult
        # start = datetime.now(timezone.utc)

        if hasattr(snapshot, '_prefetched_objects_cache') and 'archiveresult_set' in snapshot._prefetched_objects_cache:
            archive_results = [
                result
                for result in snapshot.archiveresult_set.all()
                if result.status == "succeeded" and result.output
            ]
        else:
            archive_results = snapshot.archiveresult_set.filter(status="succeeded", output__isnull=False)

        # import ipdb; ipdb.set_trace()
        link = snapshot.as_link()
        path = link.archive_path
        canon = link.canonical_outputs()
        output = ""
        output_template = '<a href="/{}/{}" class="exists-{}" title="{}">{}</a> &nbsp;'
        icons = {
            "singlefile": "❶",
            "wget": "🆆",
            "dom": "🅷",
            "pdf": "📄",
            "screenshot": "💻",
            "media": "📼",
            "git": "🅶",
            "archive_org": "🏛",
            "readability": "🆁",
            "mercury": "🅼",
            "warc": "📦"
        }
        exclude = ["favicon", "title", "headers", "htmltotext", "archive_org"]
        # Missing specific entry for WARC

        extractor_outputs = defaultdict(lambda: None)
        for extractor, _ in ArchiveResult.EXTRACTOR_CHOICES:
            for result in archive_results:
                if result.extractor == extractor and result:
                    extractor_outputs[extractor] = result

        for extractor, _ in ArchiveResult.EXTRACTOR_CHOICES:
            if extractor not in exclude:
                existing = extractor_outputs[extractor] and extractor_outputs[extractor].status == 'succeeded' and extractor_outputs[extractor].output
                # Check filesystsem to see if anything is actually present (too slow, needs optimization/caching)
                # if existing:
                #     existing = (Path(path) / existing)
                #     if existing.is_file():
                #         existing = True
                #     elif existing.is_dir():
                #         existing = any(existing.glob('*.*'))
                output += format_html(output_template, path, canon[f"{extractor}_path"], str(bool(existing)),
                                             extractor, icons.get(extractor, "?"))
            if extractor == "wget":
                # warc isn't technically it's own extractor, so we have to add it after wget
                
                # get from db (faster but less thurthful)
                exists = extractor_outputs[extractor] and extractor_outputs[extractor].status == 'succeeded' and extractor_outputs[extractor].output
                # get from filesystem (slower but more accurate)
                # exists = list((Path(path) / canon["warc_path"]).glob("*.warc.gz"))
                output += format_html(output_template, path, canon["warc_path"], str(bool(exists)), "warc", icons.get("warc", "?"))

            if extractor == "archive_org":
                # The check for archive_org is different, so it has to be handled separately

                # get from db (faster)
                exists = extractor in extractor_outputs and extractor_outputs[extractor] and extractor_outputs[extractor].status == 'succeeded' and extractor_outputs[extractor].output
                # get from filesystem (slower)
                # target_path = Path(path) / "archive.org.txt"
                # exists = target_path.exists()
                output += '<a href="{}" class="exists-{}" title="{}">{}</a> '.format(canon["archive_org_path"], str(exists),
                                                                                            "archive_org", icons.get("archive_org", "?"))

        result = format_html('<span class="files-icons" style="font-size: 1.1em; opacity: 0.8; min-width: 240px; display: inline-block">{}<span>', mark_safe(output))
        # end = datetime.now(timezone.utc)
        # print(((end - start).total_seconds()*1000) // 1, 'ms')
        return result

    cache_result = cache.get(cache_key)
    if cache_result:
        return cache_result
    
    fresh_result = calc_snapshot_icons()
    cache.set(cache_key, fresh_result, timeout=60 * 60 * 24)
    return fresh_result

   
